static char rcsid[] = "@(#)$Id: mailmsg2.c,v 2.51 2016/10/26 18:53:05 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.51 $   $State: Exp $
 *
 *  Modified by: Kari Hurtta <hurtta+elm@siilo.FMI.FI> 
 *                       (was hurtta+elm@posti.FMI.FI, hurtta+elm@ozone.FMI.FI)
 *           or  Kari Hurtta <elm@elmme-mailer.org>
 ******************************************************************************
 *  Based on Elm 2.4 src/mailmsg2.c. That code was following copyright:
 *
 *  The Elm Mail System 
 *
 * 			Copyright (c) 1988-1992 USENET Community Trust
 * 			Copyright (c) 1986,1987 Dave Taylor
 *****************************************************************************
 *  Incorparated Elm 2.5 code from src/sndmsg.c. 
 *  That code was following copyright:
 *
 *  The Elm Mail System
 *
 * This file and all associated files and documentation:
 *                      Copyright (c) 1988-1995 USENET Community Trust
 *			Copyright (c) 1986,1987 Dave Taylor
 *****************************************************************************/

/** Interface to allow mail to be sent to users.  Part of ELM  **/

#include "def_sndmail.h"
#include "s_elm.h"
#include "s_me.h"

#ifdef I_SYSEXITS
#include <sysexits.h>
#endif

DEBUG_VAR(Debug,__FILE__,"mail");

#include "menu2.h"

static char *us2s P_((unsigned char *str));
static char *us2s(str) 
     unsigned char *str;
{
    return (char *)str;
}

static unsigned char *s2us P_((char *str));
static unsigned char *s2us(str) 
     char *str;
{
    return (unsigned char *)str;
}

static const unsigned char *cs2cus P_((const char *str));
static const unsigned char *cs2cus(str) 
     const char *str;
{
    return (const unsigned char *)str;
}

#include <errno.h>
#ifndef ANSI_C
extern int errno;
#endif       

extern char version_buff[];

int  gotten_key;

static enum verify_trans_result 
  verify_transmission P_((char *filename,
			  int already_has_text, 
			  struct copy_file *copy_file,
			  int force_cmd, int options, 
			  int *dsn, 
			  struct mailing_headers *headers,
			  struct Attachments *attachments,
			  charset_t file_set,
			  struct mailer_info  *mailer_info,
			  charset_t hdr_charset,
			  struct MailboxView *mailbox,
			  int encode_hdr,
			  struct AliasView *aview,
			  struct menu_context *page,
			  struct menu_context *prompt_area,
			  struct header_rec *parent_message,
			  const struct remote_server * remote_server,
			  int *PGP_status,
			  struct elm_commands *snd_cmds));  

static void mail_sent P_((union any_fd fd, char * title, struct run_state *rs,
 			  int ret, int exit_stat));

static int append_sig P_((FILE *file,struct mailing_headers *headers,
			  struct menu_context *page));


static void mailing_message P_((int background));
static void mailing_message(background)
     int background;
{
    if (!background)
	lib_transient(CATGETS(elm_msg_cat, ElmSet, ElmSendingMail,
			      "Sending mail..."));
    else
	lib_error(CATGETS(elm_msg_cat, ElmSet, 
			  ElmSendingMailBackground,
			  "Sending mail... in background"));
    
    FlushBuffer();
}

static const unsigned char * csUs P_((const char *str));
static const unsigned char * csUs(str)
     const char *str;
{
    return (const unsigned char *)str;
}

struct string * gen_From_buffer(current_header)
     const struct header_rec * current_header;
{
    struct string * From_buffer = NULL;
    
    if (current_header->from) {

	int idx;
	int addr_item_count = 
	    addr_list_item_count(current_header->from);

	for (idx = 0; idx < addr_item_count; idx++) {

	    int group = -1;   /* Group phrases are not allowed on
				 From: -header field              */
	    const struct address * address = 
		addr_list_get_item(current_header->from,idx,&group);

	    const char          * addr     = address_get_ascii_addr(address);
	    const struct string * fullname = address_get_phrase(address);

	    if (From_buffer)
		add_ascii_to_string(From_buffer,s2us(", "));
	    else
		From_buffer = new_string(display_charset);

	    if (fullname && string_len(fullname)) {

		append_string(&From_buffer,fullname,1);

	    } else if (addr) {

		add_ascii_to_string(From_buffer,csUs(addr));

	    }
	}
    }

    return From_buffer;
}


enum mailer_capab get_mailer_level(mailer_info)
     struct mailer_info *mailer_info;
{
    enum allow_no_encoding_v an = 
	give_dt_enumerate_as_int(&allow_no_encoding);

    enum mailer_capab result = mailer_7bit;

    if (query_mailer_info(mailer_info,MI_HAVE_8BITMIME)) {
	DPRINT(Debug,8,(&Debug,"mailer supports 8BITMIME\n"));
	result        = mailer_8bit;
	
	if (query_mailer_info(mailer_info,MI_HAVE_BINARYMIME)) {
	    DPRINT(Debug,8,(&Debug,"mailer supports BINARYMIME\n"));
	    result        = mailer_binary;
	}
    }

    if (an >= allow_pass_binary) {

	/* Just send BINARY anyway */
	result     = mailer_binary;

	DPRINT(Debug,8,(&Debug,"just send binary anyway\n"));

    }  else if (an >= allow_pass_8bit &&
		result < mailer_8bit) {
	
	/* Just send 8BIT anyway */
	result     = mailer_8bit;

	DPRINT(Debug,8,(&Debug,"just 8 bit anyway\n"));
    }


    DPRINT(Debug,8,(&Debug,"get_mailer_level=%d\n",result));

    return result;
	
}

static void handle_mailer_options  P_((struct mailer_info *mailer_info,
				       mime_send_t *mime_info,
				       int *dsn,
				       struct Attachments *attachments,
				       int *reask));

static void handle_mailer_options(mailer_info,mime_info,dsn, attachments,
				  reask)
     struct mailer_info *mailer_info;
     mime_send_t *mime_info;
     int *dsn;
     struct Attachments *attachments;
     int *reask;  /* on reask loop */
{
    int i;
    int changed  = 0;

    if (query_mailer_info(mailer_info,MI_HAVE_DSN)) {
	DPRINT(Debug,8,(&Debug,"mailer supports DSN\n"));
	if(DSN_success && !reask) {
	    (*dsn) |= DSN_SUCCESS|DSN_FAILURE|DSN_DELAY;
	    DPRINT(Debug,8,(&Debug," dsn-success is set .. enabling\n"));
	}
    } else if (*dsn) {
	*dsn = 0;
	
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmNoDSN,
			  "Mailer do not support DSN, disabling it"));

	if (reask)
	    *reask = 1;

    }

    mime_info->raw_level    = get_mailer_level(mailer_info);
    
    for (i = 0; i < attachments->attachment_count; i++) {
	mime_t *part = & (attachments->attach_files[i]);
	
	if (ENCODING_BINARY == part->encoding &&
	    mime_info->raw_level < mailer_binary) {
	    part->encoding = ENCODING_BASE64;
	    changed++;
	}
		
	if (ENCODING_8BIT == part->encoding &&
	    mime_info->raw_level < mailer_8bit) {
	    part->encoding = ENCODING_BASE64;
	    changed++;
	}
    }

    if (changed) {
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmChangedAttachEnc,
			  "Changed encoding of %d attachments"),
		  changed);
	
	if (reask)
	    *reask = 1;
    }
}

static int mmsg_copy_part P_((FILE * mailbox_file,mime_t *mime_rec,
			      FILE *outfile,int quote_l, int *need_enc,
			      char **fname));

static int mmsg_copy_part(mailbox_file,mime_rec,outfile,quote_l,
			     need_enc, fname)
     FILE * mailbox_file;
     mime_t *mime_rec;
     FILE *outfile;   /* MAY be NULL */
     int quote_l;
     int *need_enc;/* MAY be NULL */
     char **fname;    
{
    struct in_state * state_in = new_in_state(STATE_in_file);
    struct in_state *decoded_state_in = NULL;
    int is_text = 0;
    int flags;
    int ret = 0;

    const char *nameref = NULL;

    if (fname)
	*fname = NULL;

    if (0 != fseek(mailbox_file,mime_rec->offset,SEEK_SET)) {
	DPRINT(Debug,3,(&Debug,
			"mmsg_copy_part: Seek failed to %ld\n",
			mime_rec->offset));
	
	goto FAIL0;
    }

    if (quote_l) {
	DPRINT(Debug,11,(&Debug,
			     "mmsg_copy_part: quote_l set, assuming textual encoding (%s)\n",
			     ENCODING(mime_rec->encoding)));

	is_text = 1;

    } else 
	is_text = decoder_is_textual(mime_rec);

    decoded_state_in = new_in_state(STATE_in_decode);

    set_in_state_file(mailbox_file,state_in);

    flags = STATE_DECODE_buffer;

    if (fname)
	flags |= STATE_DECODE_keep_file;

    if (is_text > 0)
	flags |=  STATE_DECODE_is_text;


    if (set_in_state_decode(state_in,decoded_state_in,
			    mime_rec->length,
			    flags,
			    mime_rec->encoding,
			    fname ? &nameref : NULL)) {
	int errors;

	if (outfile) {
	    int count = 0;
	    int ch;
	    int nl = 1;	    	   

	    while (EOF != (ch  = state_getc(decoded_state_in))) {

		if (nl && '[' == ch && quote_l)
		    putc('[',outfile);
		
		putc(ch,outfile);
		nl = '\n' == ch;

		count++;
	    }

	    DPRINT(Debug,11,(&Debug,
			     "mmsg_copy_part: %d characters copied to outfile\n",count));
	}

	ret = 1;
	
	/* Print total 10 error messages */
	for (errors = 0; errors < 10; errors++) {
	    const struct string *X =  
		in_state_error_message(decoded_state_in,
				       1  /* Clear error */);
	    
	    if (!X)
		break;
	    
	    lib_error(FRM("%S"),X);

	    ret = 0;
	}

	if (!ret)
	    goto FAIL1; 

	if (need_enc) {
	    if (0 != in_state_fseek(decoded_state_in,0L)) {
		 const struct string *X =  in_state_error_message(decoded_state_in,
								  1  /* Clear error */);

		DPRINT(Debug,3,(&Debug,
				"mmsg_copy_part: Seek failed to 0 (STATE_in_decode)\n"));
		
		if (X)
		    lib_error(FRM("%S"),X);

		goto FAIL1;
	    }
	    
	    *need_enc = in_state_needs_encoding(decoded_state_in,NULL,0);

	    
	    DPRINT(Debug,11,(&Debug,
			     "mmsg_copy_part: *need_enc=%d\n",*need_enc));
	}

	if (fname)
	    *fname = safe_strdup(nameref);

	/* nameref is free'ed on free_in_state(&decoded_state_in) */

    } else {

    FAIL1:
	ret = 0;
	
	/* FAILURE */
	
	if (fname) {
	    if (0 == unlink(nameref)) {
		DPRINT(Debug,11,(&Debug,
				 "mmsg_copy_part: %s unlinked\n",nameref));
	    }
	}
    }
 
    free_in_state(&decoded_state_in);

 FAIL0:
    free_in_state(&state_in);

    if (fname && *fname) {
	DPRINT(Debug,11,(&Debug,
			 "mmsg_copy_part: *fname=%s\n",*fname));
    }

    DPRINT(Debug,11,(&Debug,
			"mmsg_copy_part=%d\n",ret));
    

    return ret;
}


static int convert_attach_message P_((FILE *infile,
				      struct header_rec *current_header,
				      enum mailer_capab  mailer_level,
				      FILE * outfile,
				      enum encoding      * encoding
				      ));
static int convert_attach_message(infile,current_header,mailer_level,outfile,
				  encoding)
     FILE *infile;
     struct header_rec *current_header;
     enum mailer_capab  mailer_level;
     FILE                   * outfile;
     enum encoding          * encoding;
{
    int r;
    struct out_state  * target = new_out_state(STATE_out_file);

    set_out_state_file(outfile,target);

    r = convert_remail_message_1(infile,current_header,mailer_level,
				 target,encoding,NULL);

    /* Does not close file */
    free_out_state(&target);

    DPRINT(Debug,8,(&Debug, 
		    "convert_attach_message=%d\n",
		    r)); 

    return r;
}

int attach_rfc822(attachments,hdr,mailbox_file,title,
		  mailer_level)
     struct Attachments *attachments;
     struct header_rec  *hdr;
     FILE               *mailbox_file;
     struct string      *title;
     enum mailer_capab  mailer_level;
{
    mime_t attach;
    static int count = 0;
    const char * tmp = give_dt_estr_as_str(&temp_dir_e,"tmpdir",NULL,NULL);
    char * filename;
    FILE * tmpfp;
    enum encoding encoding = ENCODING_7BIT;

    if (!tmp)
	return 0;
    
    filename = elm_message(FRM("%selmmsg.%d-%d"), 
			   tmp, getpid (), count++);
    
    tmpfp = safeopen_rdwr(filename);
    
    if (!tmpfp) {
	free(filename);
	return 0;
    }
     
    if (hdr->status  & MIME_MESSAGE) {

	enum encoding encoding1 = ENCODING_NONE;

	if (0 != mime_warnings(hdr)) {
	    
	    DPRINT(Debug,4,(&Debug, 
			    "attach_rfc822: Bad-MIME message .. remailing as is\n"));
	    goto as_is;
	}

	DPRINT(Debug,4,(&Debug, 
			"attach_rfc822: Doing possible mime conversions\n"));
	
	/* Do possible conversions   */

	if (!convert_attach_message(mailbox_file,hdr,
				    mailer_level,tmpfp,
				    &encoding1)) {

	    unlink(filename);
	    free(filename);
	    fclose(tmpfp);
	    
	    return 0;
	}

	/* Pick up the encoding from the message. */
	update_encoding (&encoding, encoding1);

    } else {
   		
	DPRINT(Debug,4,(&Debug, 
			"attach_rfc822: Non-MIME message .. remailing as is\n"));
    as_is:
       
	if (!copy_message_f(mailbox_file,hdr,tmpfp,
			    CM_REMOVE_ENVELOPE,NULL,
			    NULL)) {
	    
	    unlink(filename);
	    free(filename);
	    fclose(tmpfp);
	    
	    return 0;
	}

	/* Pick up the encoding from the message. */
	update_encoding (&encoding,
			 hdr->mime_rec.encoding);
    }


    mime_t_zero(&attach);
    attach.pathname0 = filename;
    attach.dispname  = new_string2(system_charset,s2us(filename));
		    
    attach.length = fsize(tmpfp);
    fclose (tmpfp);
    
    attach.unlink = 2; /* mark for later deletion */
    attach.TYPE = give_media_type2(MIME_TYPE_MESSAGE,"rfc822", 0);
    if (!attach.TYPE)
	mime_panic(__FILE__,__LINE__,"attach_rfc822",
		   "message/rfc822 is not known");
 
    if (title)
	attach.description = dup_string(title);

    attach.encoding = encoding;
    add_Attachments(attachments,&attach);

    return 1;
}

int attach_body(attachments,hdr,mailbox_file,mailer_level)
     struct Attachments *attachments;
     struct header_rec  *hdr;
     FILE               *mailbox_file;
     enum mailer_capab  mailer_level;
{

    int need_enc = 0;
    mime_t attach;
    
    mime_t_zero(&attach);

    if (! mmsg_copy_part(mailbox_file,& (hdr->mime_rec),
			 NULL,0,&need_enc,
			 &attach.pathname0)) {
	
	DPRINT(Debug,4, (&Debug, 
			 "Failed to copy message \n"));
	return 0;
    }

    attach.dispname = new_string2(display_charset,
				  s2us(attach.pathname0));
    
    attach.unlink = 2; /* mark for later deletion */
    attach.TYPE = hdr->mime_rec.TYPE;
    attach.encoding     = hdr->mime_rec.encoding;
    attach.disposition  = hdr->mime_rec.disposition;
    
    if (attach.encoding >= ENCODING_NONE &&
	attach.encoding <= ENCODING_BINARY) {
	enum encoding  UNUSED_VAROK old = attach.encoding;

	if (need_enc & HAVE_BINARY) {
	    if (mailer_level >= mailer_binary)
		attach.encoding = ENCODING_BINARY;
	    else
		/* no -BBINARYMIME option */
		attach.encoding = ENCODING_QUOTED;
	}
	else if (need_enc & HAVE_8BIT) {
	    if (mailer_level >= mailer_8bit)
		/* Just send 8BIT anyway */
		attach.encoding = ENCODING_8BIT;
	    else
		/* no -B8BITMIME option */
		attach.encoding = ENCODING_QUOTED;
	}

	DPRINT(Debug,12,(&Debug,
			 "attach_body: %s need_enc=%d%s%s fixup encoding %d (%s) => %d (%s)\n",
			 attach.pathname0,
			 need_enc,
			 (need_enc & HAVE_BINARY) ? " BINARY" : "",
			 (need_enc & HAVE_8BIT)   ? " 8BIT"   : "",
			 old,ENCODING(old),
			 attach.encoding, ENCODING(attach.encoding)));	
    }



    if (hdr->mime_rec.description) {
	attach.description = 
	    dup_string(hdr->mime_rec.description);
    }
    
    if (hdr->mime_rec.TYPE_opts) {
	attach.TYPE_opts = 
	    copy_mime_param(hdr->mime_rec.TYPE_opts);
    }
    
    if (hdr->mime_rec.DISPOSITION_opts) {
	attach.DISPOSITION_opts = 
	    copy_mime_param(hdr->mime_rec.DISPOSITION_opts);
    }
    
    add_Attachments(attachments,&attach);
    
    return 1;
}

static int attach_part P_((struct Attachments *attachments,
			   struct header_rec  *hdr,
			   FILE               *mailbox_file,
			   int x,
			   enum mailer_capab  mailer_level));
static int attach_part(attachments,hdr,mailbox_file,x,mailer_level)
     struct Attachments *attachments;
     struct header_rec  *hdr;
     FILE               *mailbox_file;
     int x;
     enum mailer_capab  mailer_level;
{    
    mime_t *z = NULL;
    mime_t attach;
    int need_enc = 0;

    z = mime_parser_index(hdr->mime_rec.parser_data,x);
    if (!z) {
	DPRINT(Debug,4, (&Debug, "Failed to get part %d\n",x));
	return 0;
    }

    if (!z->TYPE) {
	DPRINT(Debug,4, (&Debug, "No type on part %d\n",x));
	return 0;
    }

    mime_t_zero(&attach);


    if (! mmsg_copy_part(mailbox_file,z,NULL,0,
			 &need_enc, &attach.pathname0)) {

	DPRINT(Debug,4, (&Debug, 
			 "Failed to copy part %d/%d\n",x));
	return 0;
    }

    attach.dispname = new_string2(display_charset,s2us(attach.pathname0));

    attach.unlink = 2; /* mark for later deletion */
    attach.TYPE = z->TYPE;
    attach.encoding     = z->encoding;
    attach.disposition  = z->disposition;

    if (attach.encoding >= ENCODING_NONE &&
	attach.encoding <= ENCODING_BINARY) {
	enum encoding  UNUSED_VAROK old = attach.encoding;
	
	if (need_enc & HAVE_BINARY) {
	    if (mailer_level >= mailer_binary)
		attach.encoding = ENCODING_BINARY;
	    else
		/* no -BBINARYMIME option */
		attach.encoding = ENCODING_QUOTED;
	}
	else if (need_enc & HAVE_8BIT) {
	    if (mailer_level >= mailer_8bit)
		/* Just send 8BIT anyway */
		attach.encoding = ENCODING_8BIT;
	    else
		/* no -B8BITMIME option */
		attach.encoding = ENCODING_QUOTED;
	}
	
	DPRINT(Debug,12,(&Debug,
			 "attach_part: %s need_enc=%d%s%s fixup encoding %d (%s) => %d (%s)\n",
			 attach.pathname0,
			 need_enc,
			 (need_enc & HAVE_BINARY) ? " BINARY" : "",
			 (need_enc & HAVE_8BIT)   ? " 8BIT"   : "",
			 old,ENCODING(old),
			 attach.encoding, ENCODING(attach.encoding)));	
    }
			
    if (z->description) {
	attach.description = dup_string(z->description);
    }

    if (z->TYPE_opts) {
	attach.TYPE_opts = copy_mime_param(z->TYPE_opts);
    }

    if (z->DISPOSITION_opts) {
	attach.DISPOSITION_opts = copy_mime_param(z->DISPOSITION_opts);
    }

    add_Attachments(attachments,&attach);

    return 1;
}

/* Return 1 if mail is saved to canceled mail */

static int cancel_ticket_to_canceled_mail 
  P_((struct mailer_cancel_ticket *ticket,
     struct mailing_headers * headers));
static int cancel_ticket_to_canceled_mail(ticket,headers)
     struct mailer_cancel_ticket *ticket;
     struct mailing_headers * headers;
{
    struct MailboxView *cm = give_canceled_mail();

    char * lbuf = NULL;
    FILE *F = NULL;
    time_t now = time(NULL);
    
    const char * s_fname = have_cancel_filename(ticket);
    FILE *       s_F     =  have_cancel_FILE(ticket);
    int ok = 0;
    int lines = -1;

    if (s_F) {
	int c;
	lines = 0;

	while (EOF != (c = getc(s_F))) {
	    if ('\n' == c)
		lines++;
	}
	rewind(s_F);

	DPRINT(Debug,10,(&Debug, "Canceled mail have %d lines.\n",
			 lines));
    }

    if (cm)
	F = add_canceled_mail(&lbuf,headers,cm,now,lines,NULL,
			      -1L);

    if (!F) {
	if (lbuf)
	    free(lbuf);
	
	lbuf = elm_message(FRM("%s/%s"), 
			   home, dead_letter);
	
	F = safeopen_rdwr(lbuf);
    } else 
	ok = 1;

    if (NULL != F) {	
	DPRINT(Debug,4,(&Debug, "Writing canceled mail to %s\n",
			lbuf));

	/* rename replaces just created file */

	if (s_fname && 0 == rename(s_fname,lbuf)) {
	    DPRINT(Debug,4,(&Debug, 
			    "Renamed %s => %s\n",
			    s_fname,lbuf));
	} else if (s_F) {
	    int c;

	    while (EOF != (c = getc(s_F))) {
		putc(c,F);
	    }
	    
	    if (0 == fflush(F) &&
		!ferror(s_F) && 
		!ferror(F)) {
		DPRINT(Debug,4,(&Debug, 
				"Copied %s => %s\n",
				s_fname,lbuf));
	    } else
		ok = 0;
	} else
	    ok = 0;
	
    }

    if (F)
	fclose(F);

    if (ok) {
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmVfyMessageKept1,
			  "Message kept.  Can be restored at next m)ail command."));
	
	if (lbuf) {
	    
	    if (last_canceled_mail)
		free(last_canceled_mail);
	    last_canceled_mail = lbuf;	    
	}
    } else if (lbuf) {
	free(lbuf);
    }
   
    return ok;
}

static FILE * generate_edit_buffer P_((struct text_block *body,
				       int * Exstat,
				       int * already_has_text,
				       charset_t * cur_editcharset));
static FILE * generate_edit_buffer(body,Exstat,already_has_text,cur_editcharset)
     struct text_block *body;
     int * Exstat;
     int * already_has_text;
     charset_t * cur_editcharset;
{
    const char * tmp = give_dt_estr_as_str(&temp_dir_e,"tmpdir",NULL,NULL);
    FILE * reply = NULL;

    if (!tmp)
	return NULL;

    /** first generate the temporary filename **/
    
    elm_sfprintf(cur_editfile,sizeof cur_editfile,
		 FRM("%s%s%d"), 
		 tmp, temp_file, getpid());
    
    /** if we're not retransmitting, create the file.. **/

    if ((reply = safeopen_rdwr(cur_editfile)) == NULL) {
	int err UNUSED_VAROK = errno;
	DPRINT(Debug,1,(&Debug,  
			"Attempt to write to temp file %s failed with error %s (mail)\n",
			cur_editfile, strerror(err)));
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCouldNotCreateFile,
			  "Could not create file %s (%s)."),
		  cur_editfile, strerror(err));

#ifdef  EX_CANTCREAT
	*Exstat = EX_CANTCREAT;  /* Give better error status */
#endif
	
	goto fail;	
    }
   
    (void) elm_chown(cur_editfile, userid, groupid);

    if (body) {
	if (! add_block_to_mail(body,
				already_has_text,
				reply,
				cur_editcharset)) {

	    DPRINT(Debug,1,(&Debug,  
			    "Failed to add body block to editfile\n"));
	    
	    goto fail;
	}
    }


    if (!reply) {
    fail:

	if (reply) {
	    fclose(reply);
	    reply = NULL;

	    unlink(cur_editfile);
	}
    }

    DPRINT(Debug,8,(&Debug,"generate_edit_buffer %s, filename %s\n",
		    reply ? "succeed" : "failed",
		    cur_editfile));

    return reply;
}

/* Returns 0 if copying of message failed */
int copy_message_to_buffer P_((FILE               * reply,
			       mime_send_t        * MIME_info,
			       charset_t          * cur_editcharset,
			       int                * already_has_text,
			       struct Attachments * attachments,
			       struct mailing_headers * headers,
			       struct MailboxView * W,
			       int                  Wi,
			       int options));

/* Returns 0 if copying of message failed */
int copy_message_to_buffer(reply,MIME_info,cur_editcharset,
			   already_has_text,attachments,
			   headers,
			   W,Wi,options)
     FILE               * reply;
     mime_send_t        * MIME_info;
     charset_t          * cur_editcharset;
     int                * already_has_text;
     struct Attachments * attachments;
     struct mailing_headers * headers;
     struct MailboxView * W;
     int                  Wi;
     int options;
{
    int forwarding     = 0 != (options & MAIL_FORWARDING);
    int mime_attach    = 0 != (options & MAIL_MIME_ATTCH);
    int copy_msg       = 0 != (options & MAIL_COPY_MSG);
    int edit_message   = 0 != (options & MAIL_EDIT_MSG);
    int delete_cancel  = 0 != (options & MAIL_DELETE_CANCEL);
    int copy_selection UNUSED_VAROK = 0 != (options & MAIL_COPY_SELECTION); 


    struct header_rec *  hdr = NULL;
    FILE              *  mailbox_file = NULL;              
    int r = 0;

    if (!give_message_data(W,Wi,
			   &hdr,&mailbox_file,NULL,
			   mime_parse_routine)) {
	
	DPRINT(Debug,4,(&Debug, 
			"Failed to retrieve original mail for replying/forwarding/canceled mail\n"));
	
	return 0;
    }

    if (forwarding || mime_attach || copy_msg) { 

	if (!*cur_editcharset)
	    *cur_editcharset = display_charset;

	if (!mime_attach) {
	    
	    if (edit_message && mailbox_file) {

		int NOHDR = forwarding ? noheaderfwd : noheader;
		int NOQUOTE = forwarding && !quote_forward; 
		int FORW = forwarding;

		r = copy_message_f(mailbox_file,hdr,
				   reply,
				   ( NOQUOTE ? 0 : CM_PREFIX ) |
				   ( NOHDR ? CM_REMOVE_HEADER : 0 ) |
				   CM_REMOVE_ENVELOPE | CM_DECODE |
				   ( FORW ? CM_FORWARDING : 0 ) |
				   ( mime_body_keywords ? CM_QUOTE_L : 0 ) |
				   CM_ATTRIBUTION | CM_FILT_HDR, 
				   *cur_editcharset,
				   NULL);

	    } else if (mailbox_file) {
		
		int NOHDR = forwarding ? noheaderfwd : noheader;
		int FORW = forwarding;

		r = copy_message_f(mailbox_file,hdr,
				   reply,
				   ( NOHDR ? CM_REMOVE_HEADER : 0 ) |
				   CM_REMOVE_ENVELOPE | CM_DECODE |
				   ( FORW ? CM_FORWARDING : 0 ) |
				   ( mime_body_keywords ? CM_QUOTE_L : 0 ) |
				   CM_ATTRIBUTION, *cur_editcharset,
				   NULL);
	    }

	    if (r)
		*already_has_text = 1;

	} else { /* mime_attach */

	    struct string * title = NULL;
	    struct string * From_buffer = gen_From_buffer(hdr);	    

	    if (forwarding) {
		if (From_buffer)
		    title = format_string(CATGETS(elm_msg_cat, ElmSet, 
						  ElmForwardedMesg1,
						  "Forwarded message from %S"), 
					  From_buffer);
		else
		    title =	format_string(CATGETS(elm_msg_cat, ElmSet, 
						      ElmForwardedMesg2,
						      "Forwarded message from %s"), 
					      hdr->env_from);
	    }

	    r = attach_rfc822(attachments, hdr,mailbox_file,title,
			      MIME_info->raw_level);

	    if (title)
		free_string(&title);
	    
	    if (From_buffer)
		free_string(&From_buffer);	       	    

	    if (!r) {
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFailFileForward,
				  "Failed to create file for forwarding"));

	    }

	}

    } else {  /* not forwarding, mime_attach, copy_msg */

	/* make referenced message as edited message */
	header_list_ptr all_headers,walk;
	int Flags, major;
	const char *subtype;

	DPRINT(Debug,4, (&Debug,
			 "Not forwarding, copying message normally or mime attaching, making message as edited -- copy_selection=%d\n",
			 copy_selection));
	
	if (skip_envelope(hdr,mailbox_file) == -1) {
	    DPRINT(Debug,4, (&Debug, "Can't skip envelope of message \n"));
	    
	    return 0;
	}

	all_headers = file_read_headers(mailbox_file,RHL_CHECK_HEADER);

	for (walk = all_headers; walk; walk = walk -> next_header) {
	    const char * hdr_name = give_header_name(walk->header_name);
	    
	    if (0 == strincmp(hdr_name,"Content-",8) ||
		0 == istrcmp(hdr_name,"MIME-Version") ||
		    0 == istrcmp(hdr_name,"Received") ||
		0 == istrcmp(hdr_name,"Sender"))
		continue;
	    
	    if (!add_to_mailing_header(headers,walk->header_name,
				       walk->body,
				       !(hdr->status & NOHDRENCODING),
				       hdr->header_charset,1)) {
		DPRINT(Debug,4, (&Debug, "Header %s not copied\n",
				 hdr_name));
	    }
	}
	
	dump_expanded_address(4,"after copy -- from ",headers->from);
	dump_expanded_address(4,"after copy -- to   ",headers->to);
	dump_expanded_address(4,"after copy -- cc   ",headers->cc);

	if (! hdr->mime_rec.TYPE) {
	    DPRINT(Debug,4, (&Debug, "no mime type\n"));
	    goto fail1;		
	}

	Flags   =  get_type_flags(hdr->mime_rec.TYPE);
	major   =  get_major_type_code(hdr->mime_rec.TYPE);
	subtype =  get_subtype_name(hdr->mime_rec.TYPE);

	/* 1) Includes only one text part -- put it to edit buffer */
	if (MIME_TYPE_TEXT == major && 0 == istrcmp(subtype,"Plain") &&
	    DISP_INLINE == hdr->mime_rec.disposition) {
	    
	    const char *cs = NULL;
	    
	    if (! mmsg_copy_part(mailbox_file,& (hdr->mime_rec),
				 reply,mime_body_keywords,NULL,NULL)) {
		
		DPRINT(Debug,4, (&Debug, "Failed to copy part\n"));
		
		goto fail1;
	    }
	    
	    cs = get_mime_param_compat(hdr->mime_rec.TYPE_opts,
				       "charset");
	    if (!cs) {
		DPRINT(Debug,4, (&Debug, "No charset parameter\n"));

		/* Not necessary correct */
		*cur_editcharset = display_charset;

	    } else {

		*cur_editcharset = MIME_name_to_charset(cs,0);
		if (!*cur_editcharset) {
		    DPRINT(Debug,4, (&Debug, "charset parameter %s unknown?\n",
				     cs));
		    goto fail1;
		}
		
		if (charset_superset_of(display_charset,*cur_editcharset)) {
		    const char *MIME_name_c UNUSED_VAROK = 
			get_charset_MIME_name(*cur_editcharset);
		    const char *MIME_name_d UNUSED_VAROK = 
			get_charset_MIME_name(display_charset);
		    
		    DPRINT(Debug,4, (&Debug, 
				     "charset parameter %s (%s) is subset of display charset %s\n",
				     MIME_name_c ? 
				     MIME_name_c : 
				     "<no MIME name>",
				     cs,
				     MIME_name_d ?
				     MIME_name_d :
				     "<no MIME name>"));
		    *cur_editcharset = display_charset;
		}
	    }
	    
	    *already_has_text = 1;
	    r = 1;
	    DPRINT(Debug,4, (&Debug, "Added whole mail to editing buffer\n"));
	    
	    /* 2) multipart MIXED */
	} else if ((Flags & MIME_MIXED) &&
		   DISP_INLINE == hdr->mime_rec.disposition &&
		   hdr->mime_rec.parser_data) {
	    
	    int s = 0;
	    int c = mime_parser_subparts(hdr->mime_rec.parser_data);
	    mime_t *z = NULL;
	    int x;
	    
	    /* Pick first parts to editing buffer */
	    if (c > 0 &&
		(z = mime_parser_index(hdr->mime_rec.parser_data,0)) &&
		z->TYPE) {
		int major           =  get_major_type_code(z->TYPE);
		const char *subtype =  get_subtype_name(z->TYPE);
		
		const char *cs = NULL;
		
		if (MIME_TYPE_TEXT == major && 0 == istrcmp(subtype,"Plain") &&
		    DISP_INLINE == z->disposition) {
		    
		    if (! mmsg_copy_part(mailbox_file,z,reply,
					 mime_body_keywords,
					 NULL,NULL)) {
			DPRINT(Debug,4, (&Debug, "failed to copy part\n"));
			goto fail1;
		    }
		    
		    cs = get_mime_param_compat(z->TYPE_opts,
					       "charset");
		    if (!cs) {
			DPRINT(Debug,4, (&Debug, "No charset parameter?\n"));
			goto fail1;
		    }
		    
		    *cur_editcharset = MIME_name_to_charset(cs,0);
		    if (!*cur_editcharset) {
			DPRINT(Debug,4, (&Debug, "charset parameter %s unknown?\n",
					 cs));
			goto fail1;
		    }
		    
		    if (charset_superset_of(display_charset,*cur_editcharset)) {
			
			const char *MIME_name_c UNUSED_VAROK = 
			    get_charset_MIME_name(*cur_editcharset);
			const char *MIME_name_d UNUSED_VAROK = 
			    get_charset_MIME_name(display_charset);
						    
			DPRINT(Debug,4, (&Debug, 
					 "charset parameter %s (%s) is subset of display charset %s\n",
					 MIME_name_c ?
					 MIME_name_c :
					 "<no MIME name>",
					 cs,
					 MIME_name_d ?
					 MIME_name_d :
					 "<no MIME name>"));
			*cur_editcharset = display_charset;
		    }
			
		    DPRINT(Debug,4, (&Debug, 
				     "Added first part of mail to editing buffer\n"));
		    
		    /* Start attachments from second part */
		    s = 1;		    
		}       	
	    }

	    r = 1;			
	    
	    for (x = s; x < c; x++) {
		
		if (! attach_part(attachments, hdr,mailbox_file,x,
				  MIME_info->raw_level)) {
		    DPRINT(Debug,4, (&Debug, 
				     "Failed to copy part %d/%d\n",
				     x,c));
		    continue;
		}
		
		
		DPRINT(Debug,4, (&Debug, 
				 "Added part %d/%d\n",
				 x,c));
		
	    }
	    
	    
	    /* 3) Add just whole mail as attachment */
	} else {
	    
	    if (! attach_body(attachments, hdr,mailbox_file,
			      MIME_info->raw_level)) {
		DPRINT(Debug,4, (&Debug, 
				 "Failed to copy messge \n"));
		goto fail1;
	    }
	    
	    
	    DPRINT(Debug,4, (&Debug, 
			     "Added message\n"));
	    
	    r = 1;
	}
	
	if (delete_cancel)
	    delete_current_cancel(W,mailbox_file);

	
    fail1:

	if (all_headers)
	    delete_headers(&all_headers);

    }

    if (!r) {
	DPRINT(Debug,4,(&Debug, 
			"Failed to process original mail for replying/forwarding/canceled mail\n"));
    }

    return r;
}			       

/* Returns 0 if reading of body failed */
int mail(mail_index,options, headers,mailer_info,
	 mailbox,aview, parent_page,
	 selection_view,selection_idx,
	 body,
	 exit_status,pgp_status)
     int mail_index;
     int options;
     struct mailing_headers * headers;
     struct mailer_info *mailer_info;     
     struct MailboxView *mailbox /* for mail_index */;
     struct AliasView *aview;
     struct menu_context  *parent_page;
     struct MailboxView *selection_view;
     int selection_idx;
     struct text_block *body;
     int *exit_status;               /* Only used on send only mode */
     int pgp_status;
{
    int replying      = 0 != (options & MAIL_REPLYING);
    int copy_msg      = 0 != (options & MAIL_COPY_MSG);
    int forwarding    = 0 != (options & MAIL_FORWARDING);
#if 0
    int mime_attach   = 0 != (options & MAIL_MIME_ATTCH);
#endif
    int copy_selection = 0 != (options & MAIL_COPY_SELECTION); 
    int delete_cancel  = 0 != (options & MAIL_DELETE_CANCEL);

    struct Attachments attachments = NULL_Attachments;

    /** Given the addresses and various other miscellany (specifically, 
	'copy-msg' indicates whether a copy of the current message should 
	be included, (options & MAIL_EDIT_MSG) indicates whether the message should 
	be edited) this routine will invoke an editor for the user and 
	then actually mail off the message. 

	In redraw is needed use 
	       use menu_trigger_redraw(parent_page)
    **/

    FILE *reply = NULL;
    int command_ch = '\0';

    int  already_has_text = 0;		/* we need an ADDRESS */
    int	 signature_done   = 0;
    int	 err;
    int reask_verify = 0;

    int dsn = 0;
    mime_send_t MIME_info;
    
    FILE * converted_buffer = NULL;      /* Temp file which hold converted
					    body parts -- now include mime
					    headers however ...
					 */
    static int conv_count = 0;
    char * temp1 = NULL;

    char title[80];
    int code = ft_none;
    struct copy_file COPY_FILE;

    charset_t   cur_editcharset = NULL;
    
    struct menu_context *page = NULL;
    struct menu_context *prompt_area = NULL;
    const char * tmp = give_dt_estr_as_str(&temp_dir_e,"tmpdir",NULL,NULL);
    int from_addr_len;
    int ret = 0;

    const struct remote_server  * remote_server = NULL;
    struct mailer_cancel_ticket * cancel_ticket = NULL;
    struct sending_display      * sending_display = NULL;

    int Exstat = 0;   /* OK */
    int restart_loop_count = 0;
    struct elm_commands *snd_cmds = NULL;

    if (!tmp)
	return 0;

    /* Fail if no mailer available */
    if (!mailer_info)
	return 0;

    /* Initialize structure */
    clear_mime_send_info(&MIME_info);

    /* free_mailing_headers will free this */
    if (!headers->env_from)
	headers->env_from = mailer_get_env_from(mailer_info);

    cancel_ticket = new_mailer_cancel_ticket();
    snd_cmds      =  give_send_commands();

    zero_copy_file(&COPY_FILE);

    dump_expanded_address(4,"Mailing to",headers->to);

    if (0 != (options & MAIL_EDIT_MSG)) {
	
	command_ch = 'e';
	DPRINT(Debug,4, (&Debug,"   (with editing)"));
    }
    DPRINT(Debug,4, (&Debug,"\n"));
        
    gotten_key = 0;		/* ignore previously gotten encryption key */
       
    /* verify_transmission() displays new screen,
       we create context for it on here
    */
       
    page = new_menu_context();
    if (OPMODE_IS_INTERACTIVE(opmode)) {
	int LINES, COLUMNS;

	menu_ClearScreen(page);

	menu_get_sizes(page,&LINES, &COLUMNS);

	prompt_area = 
	    new_menu_subpage(page,LINES-4,4,
			     subpage_simple_noredraw,NULL);

    }

    reply = generate_edit_buffer(body,&Exstat,
				 &already_has_text,
				 &cur_editcharset);
    if (!reply)
	goto fail_label;


    handle_mailer_options(mailer_info,&MIME_info,&dsn, 
			  &attachments, NULL);

    if (mailbox)
	remote_server = give_message_remote_server(mailbox,mail_index);

    if ((mailbox && mail_index >= 0 &&
	 (copy_msg || forwarding)) 
	|| 
	copy_selection) {
       
	struct MailboxView *  W  = mailbox;
	int                   Wi = mail_index;

	if (copy_selection) {
	    W  = selection_view;
	    Wi = selection_idx;
	}

	if (! copy_message_to_buffer(reply,&MIME_info,
				     &cur_editcharset,
				     &already_has_text,
				     &attachments,
				     headers,
				     W,Wi,options)) {
	    goto fail_label;
	}				            			     			      
    }
    
    if (!cur_editcharset)
	cur_editcharset = display_charset;

    if (attach_files.attachment_count > 0) {
	int i;

	for (i = 0; i < attach_files.attachment_count; i++) {
	    add_Attachments(& attachments,&attach_files.attach_files[i]);
	    /* add_Attachments zeros attach_files.attach_files[i],
	       so array can be fre'ed later 
	    */
	}
	free_Attachments(&attach_files); /* Actually only frees array */
    }

    /* append signature now if we are going to use an external editor */
    /* Don't worry about the remote/local determination too much */
    
    {
	unsigned int editor_keyword = 0, editor_mask = 0;
	const char * editor_val = give_dt_estr_as_str(&editor_e,"editor",
						      &editor_keyword,
						      &editor_mask);

	if (already_has_text || 
	    (editor_val  &&
	     editor_kw_NO == editor_keyword)) {
	    signature_done = TRUE;

	    already_has_text |= append_sig(reply, headers, page);
	}
    }

    /** Edit the message **/

    /* calculate default save_file name */
    if (COPY_FILE.copy_file)
	clear_copy_file(&COPY_FILE);        /* signals to not save a copy */
    
    if(auto_copy_sent) {
	if(save_by_name || save_by_alias) {
	    if(force_name) {
		/* signals save by 'to' logname */
		COPY_FILE.copy_file = format_string(FRM("="));
	    } else {
		/* conditional save by 'to' logname */
		COPY_FILE.copy_file = format_string(FRM("=?"));
	    }
	} else {
	    /* signals save to sentmail */
	    COPY_FILE.copy_file = format_string(FRM("<"));
	}
    } 
    
    if (use_PGP &&
	mailbox && mail_index >= 0) {
	struct header_rec *hdr =  give_header(mailbox,mail_index);
	if (hdr &&
	    hdr->pgp & PGP_MESSAGE)
	    options |= MAIL_HAVE_PGP_ENCODED;
    }

    /* Now .mailheaders is read before mail is composed! */
    if (!copy_selection)
	import_mailheaders(headers);

    if (0) {
    restart_verify_transmission:
	++restart_loop_count;
	DPRINT(Debug,8,(&Debug,"... restart verify_transmission, count=%d\n",
			restart_loop_count));
	if (restart_loop_count > 50) {

#ifdef  EX_UNAVAILABLE
	    Exstat = EX_UNAVAILABLE;  /* Give better error status */
#endif
	    goto fail_label;	
	    
	}
    }

 verify_trans_menu:
    do {      /* verify_transmission loop */
	
	if (sending_display) {
	    free_sending_display(&sending_display);
	    menu_trigger_redraw(page);
	}

	if (reply != NULL) {
	    int r = fclose(reply);	
	    reply = NULL;

	    if (EOF == r) {
		int err = errno;
		lib_error(CATGETS(elm_msg_cat, ElmSet, 
				  ElmWriteFailedCloseFile,
				  "Write failed when closing file %s: %s"),
			  cur_editfile, strerror(err));
			  
#ifdef  EX_IOERR
		Exstat = EX_IOERR;  /* Give better error status */
#endif
			  
		goto fail_label;
	    }
	}

	code = ft_none;
	reask_verify = 0;
	
	/* ask the user to confirm transmission of the message */
	if (OPMODE_IS_INTERACTIVE(opmode)) {
	    
	    struct header_rec *parent_message;
	    
	    parent_message = NULL;
	    
	    menu_set_default(page);
	    
	    if (replying && mailbox && mail_index >= 0) 
		parent_message =  give_header(mailbox,mail_index);
	    
	    if (command_ch) {
		if (isascii(command_ch) && 
		    isprint(command_ch)) {
		    DPRINT(Debug,8,(&Debug,"... command: '%c'\n",
				    command_ch));
		} else {
		    DPRINT(Debug,8,(&Debug,"... command: %d\n",
				    command_ch));
		}
	    }

	    code = verify_transmission(cur_editfile, 
				       already_has_text, &COPY_FILE, 
				       command_ch,
				       options, &dsn,
				       headers, &attachments,
				       cur_editcharset,
				       mailer_info,
				       MIME_info.hdr_charset,
				       mailbox,
				       MIME_info.encode_hdr,
				       aview,
				       page,
				       prompt_area,
				       parent_message,
				       remote_server,
				       &pgp_status,
				       snd_cmds);	    
	    command_ch = '\0';  /* Reset command anyway */
	    
	    DPRINT(Debug,8,(&Debug,
			    " ... verify_transmission result = %d, attachment_count=%d\n",
			    code,attachments.attachment_count));
	    
	    switch(code) {
	    case ft_forget_empty:
		DPRINT(Debug,8,(&Debug," ... ft_forget_empty\n"));
		/* Empty ... */
		if (!attachments.attachment_count) 
		    goto fail_label;	    
		break;
	    case ft_EOF:
		DPRINT(Debug,8,(&Debug," ... ft_EOF\n"));
		goto fail_label;
	    }
	} else {
	    DPRINT(Debug,8,(&Debug,"Non interactive -- assuming 'send' \n"));

	    code = ft_send;
	}

	err = can_open(cur_editfile,"r+");
	if (err) {
	    DPRINT(Debug,1,(&Debug,  
			    "Attempt to open file %s for reading failed with error %s (mail) -- can_open\n",
			    cur_editfile, strerror(err)));
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCouldNotOpenReply,
			      "Could not open reply file (%s)."), 
		      strerror(err));

	    goto fail_label;
	}

	if ((reply = fopen(cur_editfile,"r+")) == NULL) {
	    err = errno;
	    DPRINT(Debug,1,(&Debug,  
			    "Attempt to open file %s for reading failed with error %s (mail)\n",
			    cur_editfile, strerror(err)));
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCouldNotOpenReply,
			      "Could not open reply file (%s)."), 
		      strerror(err));

	    goto fail_label;
	}

	/* Append signature if not done earlier, so that need_eccoding
	 * takes account of 8-bit data on signature
	 */
	
	if (!signature_done) {
	    /* Go to the end of the file! */
	    fseek (reply, 0, 2);
	    append_sig(reply, headers, page);
	    rewind(reply);
	    signature_done = TRUE;
	}
	
	if (code >= ft_send) { /* If not canceled */
	    int r = mailer_restarted(mailer_info);

	    if (r > 0)
		handle_mailer_options(mailer_info,&MIME_info,&dsn, 
				      &attachments, &reask_verify);

	    if (r < 0) {
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmMailerDisconnected,
				  "Mailer disconnected! Can't send message. Use f)orget !"));


		reask_verify = TRUE; /* Go to verify_transmission again. */
	    }
	}

	MIME_info.encoding_top  = ENCODING_7BIT;/* Encoding for Multipart/ */
	  	 
#if 0	  
	if (allow_no_hdrencoding && headers->subject)
	    MIME_info.need_enc |= check_8bit_string (headers->subject);
#endif
	
	if (check_for_multipart(reply, &MIME_info,mailer_info,
				cur_editcharset, remote_server) < 0) { 
	    /* Error in [include ...] or other [keyword] */
	    if (code < ft_send)
		goto fail_label;

	    if (OPMODE_IS_INTERACTIVE(opmode)) {
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFixIncludeKeyword,
				  "Please fix [include ...] and other [keyword] lines!"));
		reask_verify = TRUE; /* Go to verify_transmission again. */
		continue;
	    }
	}

	if (converted_buffer)
	    fclose(converted_buffer);
	temp1 = elm_message(FRM("%selmcv-%d-%d"),
			    tmp, getpid (),
			    conv_count++);
	converted_buffer = safeopen_rdwr(temp1);
	if (!converted_buffer) {

	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmOpenFailedCopy,
			      "Temp file open failed to in copy"));


	    free(temp1);
	    temp1 = NULL;

#ifdef  EX_CANTCREAT
	    Exstat = EX_CANTCREAT; /* Give better error status */
#endif

	    if (code < ft_send ||
		! OPMODE_IS_INTERACTIVE(opmode))
		goto fail_label;
	    
	    reask_verify = TRUE; /* Go to verify_transmission again. */
	    continue;
	}
	DPRINT(Debug,2,(&Debug,   "mail: using temp file (%s)\n",
			temp1));
	unlink(temp1);
	free(temp1);
	temp1 = NULL;

	if (!convert_text(reply,converted_buffer, &MIME_info,
			  cur_editcharset,text_charset,
			  pgp_status,
			  &attachments,
			  mailer_info,
			  page,prompt_area,
			  remote_server)) {

#ifdef  EX_DATAERR
	    Exstat = EX_DATAERR;  /* Give better error status */  
#endif

	    if (code < ft_send ||
		! OPMODE_IS_INTERACTIVE(opmode))
		goto fail_label;

	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFailedConvertMessage,
			      "Failed convert message, fix errors and retry or f)orget it."));
	    reask_verify = TRUE; /* Go to verify_transmission again. */
	} else
	    MIME_info.msg_is_multipart = MIME_info.top_parts_count > 1;

	
	DPRINT(Debug,8,(&Debug,
			" ... top_parts_count=%d, msg_is_multipart=%d\n",
			MIME_info.top_parts_count,
			MIME_info.msg_is_multipart));


	/* End of check_for_multipart failure loop */
    } while (OPMODE_IS_INTERACTIVE(opmode) && reask_verify);
	
    if (MIME_info.msg_is_multipart &&
	!MIME_info.mime_boundary[0]) {
	(void) mime_generate_boundary (MIME_info.mime_boundary,
				       sizeof MIME_info.mime_boundary);
	mime_params_add_compat(&(MIME_info.TYPE_opts_top), 
			       "boundary",
			       MIME_info.mime_boundary);

	DPRINT(Debug,8,(&Debug,
			" ... mime boundary: %s\n",MIME_info.mime_boundary));
    }

    if (headers->from.addrs &&
	(from_addr_len = addr_list_item_count(headers->from.addrs)) > 0) {

	DPRINT(Debug,6,(&Debug,
			"From address already specified by user (len %d)\n",
			from_addr_len));

	dump_expanded_address(6,"From address",headers->from);

	if (from_addr_len > 1 || add_sender_header) {
	    headers->sender = from_addr_literal(mailer_info);
	    DPRINT(Debug,6,(&Debug,
			    "setting Sender address: %s\n",
			    headers->sender));
	}

    } else if (!query_mailer_info(mailer_info,MI_DONT_ADD_FROM))
	make_from_addr(&(headers->from),mailer_info);
    
    if (headers->env_from && env_from_changed(headers->env_from) &&
	add_sender_header && !headers->sender) {
	    headers->sender = from_addr_literal(mailer_info);
	    DPRINT(Debug,6,(&Debug,
			    "setting Sender address: %s\n",
			    headers->sender));	    
    }

    switch(code) {

    case ft_preview: 
    case ft_preview_hdrs:
    case ft_preview_menu: 
    case ft_preview_mime: {

	enum snd_preview_result pvr = 

	    preview_message(converted_buffer,&command_ch,
			    page,snd_cmds,code,
			    &MIME_info,headers,
			    aview,mailbox,remote_server);
	switch(pvr) {
	case sp_none: 
	    lib_error(CATGETS(elm_msg_cat, ElmSet, 
			      ElmFailedPreviewMess,
			      "Failed to preview message. Use f)orget !"));

	    if (OPMODE_IS_INTERACTIVE(opmode))
		goto restart_verify_transmission;

	    break;

	case sp_send: goto send_message;
	case sp_verify_trans: 
	    DPRINT(Debug,8,(&Debug," ... sp_verify_trans: "));

	    if (command_ch) {
		DPRINT(Debug,8,(&Debug,"command_ch=%d",
				command_ch));
	    } else {
		DPRINT(Debug,8,(&Debug,"no command"));
	    }
	    DPRINT(Debug,8,(&Debug,"\n"));
	    	    
	    goto verify_trans_menu; 
	case sp_forget: 
	    if (reply && fsize(reply) == 0) {		
		DPRINT(Debug,8,(&Debug," ... sp_forget: empty\n"));
		/* Empty ... */
		if (!attachments.attachment_count) 
		    break;
	    }
	    goto forget_message;
	case sp_EOF:
	    DPRINT(Debug,8,(&Debug," ... sp_EOF\n"));
	    break;
	}
    }
	break;
	
    case ft_forget:
    case ft_forget_empty: 
    forget_message: { /* Forget */	

	struct MailboxView *cm = give_canceled_mail();

	char * lbuf = NULL;
	struct out_state  * fp_copy = new_out_state(STATE_out_file);
	FILE *F = NULL;
	time_t now = time(NULL);

	if (cm)
	    F = add_canceled_mail(&lbuf,headers,cm,now,-1,NULL,
				  -1L);
				  
	if (!F) {
	    if (lbuf)
		free(lbuf);

	    lbuf = elm_message(FRM("%s/%s"), 
			       home, dead_letter);

	    F = safeopen_rdwr(lbuf);
	} else {
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmVfyMessageKept1,
			      "Message kept.  Can be restored at next m)ail command."));
	}
	
	if (NULL != F) {
	    
	    DPRINT(Debug,4,(&Debug, "Writing canceled mail to %s\n",
			    lbuf));
	    
	    set_out_state_file(F,fp_copy);
	    write_header_info(fp_copy,headers,cm_across_copy,
			      &MIME_info);

	    copy_message_across(&MIME_info,fp_copy,
				cm_across_copy,converted_buffer);

	}
	free_out_state(&fp_copy);
	if (F)
	    fclose(F);

	if (lbuf) {

	    if (last_canceled_mail)
		free(last_canceled_mail);
	    last_canceled_mail = lbuf;	    
	}
    }
	break;
	
    case ft_send:
    send_message: {
	int r UNUSED_VAROK;
	struct mail_send_state * mailer = NULL;
	/* Only do prompts on interactive mode */
	struct menu_context *p = 
	    OPMODE_IS_INTERACTIVE(opmode) ? prompt_area : NULL;


	rewind(converted_buffer);
	if (COPY_FILE.copy_file) /* i.e. if copy_file contains a name */
	    save_copy(headers,&COPY_FILE, &MIME_info, 
		      converted_buffer, 
		      page,aview,p);  /* Need access for all open 
				       mailboxes/folders */

	/** write all header information into whole_msg_file **/

	/** try to write headers to new temp file **/
	
	DPRINT(Debug,6,(&Debug,   "Composition file='%s'\n", 
		   cur_editfile));
	
	dump_expanded_address(2,"From",headers->from);
	dump_expanded_address(2,"To",headers->to);
	dump_expanded_address(2,"CC",headers->cc);
	dump_expanded_address(2,"Bcc",headers->bcc);

	{
	    int count = 0;

	    char **argvt = argv_from_headers(headers);

	    if (!argvt) {
		DPRINT(Debug,4,(&Debug,"mail: No mail addresses\n"));
		
		if (OPMODE_IS_INTERACTIVE(opmode))
		    goto restart_verify_transmission;
		
		goto fail_label;
	    }

	    mailer = mailer_init(argvt,dsn,sendmail_verbose,mailer_info,
				 headers->env_from);

	    while (argvt[count])
		count++;

	    free_argv(&argvt);

	    if (sending_display)
		free_sending_display(&sending_display);
	    
	    if (query_mailer_info(mailer_info,
				  MI_HAVE_SENDING_DISPLAY)) {

		DPRINT(Debug,4,(&Debug,"mail: Mailer have sending display\n"));

		sending_display = 
		    new_sending_display(count,
					OPMODE_IS_INTERACTIVE(opmode));
		
	    }
	}

	if (!mailer) {

	    if (OPMODE_IS_INTERACTIVE(opmode))
		goto restart_verify_transmission;
	    
	    goto fail_label;
	} else if (sending_display) {

	    if (! mailer_set_sending_display(mailer,
					     sending_display,
					     free_sending_display,
					     inc_sending_display_refcount,
					     update_addr_sd,
					     check_addr_sd,
					     update_data_sd,
					     check_fail_sd,
					     update_con_sd,
					     size_check_sd,
					     update_size_cb)) {
		DPRINT(Debug,4,(&Debug,
				"mail: Mailer did not supported sending display ( mailer_set_sending_display failed)\n"));
	    }
	}
	
	/* Let state routines be aware of encoding */
	if (MIME_info.encoding_top == ENCODING_BINARY)
	    set_out_state_EOLN(get_mail_outfd(mailer),1);

	write_header_info(get_mail_outfd(mailer),headers,
			  cm_across_send, &MIME_info);

	rewind(converted_buffer);

	if (!copy_message_across(&MIME_info,get_mail_outfd(mailer),
				 cm_across_send,converted_buffer)) {

	    if (OPMODE_IS_INTERACTIVE(opmode))
		goto restart_verify_transmission;
	    
	    goto fail_label;
	}
	
	if (headers->to.addrs) {
	    int cnt;
	    
	    if ((cnt = addr_list_item_count(headers->to.addrs)) == 1) {
		int group;
		const struct address * X = 
		    addr_list_get_item(headers->to.addrs,0,&group);
		const char * addr = address_get_ascii_addr(X);
		
		elm_sfprintf(title, sizeof title,
			     CATGETS(elm_msg_cat, ElmSet, ElmMailTo,
				     "Mail to %.50s..."),
			     addr ? addr : "(someone)");
		
	    } else
		elm_sfprintf(title, sizeof title,
			     CATGETS(elm_msg_cat, ElmSet, ElmMailToXRecipients,
				     "Mail to %d recipients (To:)"),
			     cnt);


	} else
	    elm_sfprintf(title, sizeof title,
			 CATGETS(elm_msg_cat, ElmSet, ElmMailToSomeone,
				 "Mail to someone..."));


	r = mail_backend3(&mailer,mail_sent, 
			  MIME_info.encoding_top, title,mailing_message,
			  &Exstat,
			  cancel_ticket,
			  /* message size */
			  MIME_info.cl_end);
	
	DPRINT(Debug,4,(&Debug,"mail_backend2 returned %d, exit status = %d\n",
			r,Exstat));

	if (cancel_ticket) {
	    enum msd_action action = have_msd_action(cancel_ticket);
	    switch (action) {
	    case msd_act_EOF: 
		DPRINT(Debug,4,(&Debug,
				"  cancel_ticket: action=msd_act_EOF (%d)\n",
				action));
		    
		code = ft_EOF;  /* LEAVE */
		goto fail_label;
		
	    case msd_act_SIG:
		DPRINT(Debug,4,(&Debug,
				"  cancel_ticket: action=msd_act_SIG (%d)\n",
				action));
		goto fail_label;
		
	    case msd_act_need_restart: {
		int r;
		
		DPRINT(Debug,4,(&Debug,
				"  cancel_ticket: action=msd_act_need_restart (%d)\n",
				action));

		r = mailer_restarted(mailer_info);
		
		if (r > 0) {
		    handle_mailer_options(mailer_info,&MIME_info,&dsn, 
					  &attachments, &reask_verify);
		    
		    /* Need rebuild mail */
		}
		
		if (r < 0) {
		    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmMailerDisconnected,
				      "Mailer disconnected! Can't send message. Use f)orget !"));
		    
		    
		    if (OPMODE_IS_INTERACTIVE(opmode))
			goto restart_verify_transmission;
		    
		    cancel_ticket_to_canceled_mail(cancel_ticket,headers);
		    goto fail_label;
		}
		
		goto restart_verify_transmission;		
	    }
		break;
	    case msd_act_default:
		DPRINT(Debug,4,(&Debug,
				"  cancel_ticket: action=msd_act_default (%d)\n",
				action));
		
		break;
	    case msd_act_cancel:
		DPRINT(Debug,4,(&Debug,
				"  cancel_ticket: action=msd_act_cancel (%d)\n",
				action));
		
		cancel_ticket_to_canceled_mail(cancel_ticket,headers);
		
		goto fail_label;

	    case msd_act_edit:
		DPRINT(Debug,4,(&Debug,
				"  cancel_ticket: action=msd_act_edit (%d)\n",
				action));

		if (OPMODE_IS_INTERACTIVE(opmode))
		    goto restart_verify_transmission;

		cancel_ticket_to_canceled_mail(cancel_ticket,headers);
		goto fail_label;

	    case msd_act_send:
		DPRINT(Debug,4,(&Debug,
				"  cancel_ticket: action=msd_act_send (%d)\n",
				action));
		break;
	    }
	}

	   

 	/* 0   may be:   start failed
 	 *               mailing continued on backend
 	 *
  	 *  1            command completed, perhaps with failure 
 	 */
	
 	ret = 1;


	/* NOTE: succees failure, backgrouded status currently ignored */

	/* mark the "replied" status of the message */
	if (replying && mailbox && mail_index >= 0 && ret) {

	    struct header_rec *hdr =  give_header(mailbox,mail_index);
	    if (hdr && 	0 == (hdr->status & REPLIED_TO)) {
		
		DPRINT(Debug,4,(&Debug,
				"Adding replied mark to current message\n"));
		
		hdr->status |= REPLIED_TO;
		hdr->status_chgd = TRUE;
	    } else {
		DPRINT(Debug,8,(&Debug,
				"Not adding replied mark %s\n",
				hdr ? "" : " -- getting header failed"));
	    }

	} else {
	    DPRINT(Debug,8,(&Debug,"Not adding replied mark %s%s%s\n",
			    replying ? "" : " -- not replying ",
			    mailbox ? "" : " -- no mailbox",
			    ret ? "" : " -- no succeed"));
	}
    }
	break;

    }
	
 fail_label:	
    if (reply)
	fclose(reply);

    if (sending_display)
	free_sending_display(&sending_display);

    if (delete_cancel)
	sync_canceled_mails(selection_view);

    if (snd_cmds)
	free_commands(&snd_cmds);

    if (cancel_ticket)
	free_mailer_cancel_ticket(&cancel_ticket);
    
    free_Attachments(&attachments);

    if (converted_buffer)
	fclose(converted_buffer);
    converted_buffer = NULL;

    /* Unlink temp file now.
     * This is a precaution in case the message was encrypted.
     * I.e. even though this file is readable by the owner only,
     * encryption is supposed to hide things even from someone
     * with root privileges. The best we can do is not let this
     * file just hang after we're finished with it.
     */

    if (cur_editfile[0]) {
	if (0 ==unlink (cur_editfile)) {
	    DPRINT(Debug,8,(&Debug,"Removing %s\n",
			    cur_editfile));
	} 
    }
    
    free_mime_send_info(&MIME_info);
    clear_copy_file(&COPY_FILE);


    if (prompt_area)
	erase_menu_context(&prompt_area);


    if (page)
	erase_menu_context(&page);

    if (parent_page)
	menu_trigger_redraw(parent_page);

    if (ft_EOF == code) {
	if (0 == Exstat) {
	    Exstat = 
#ifdef  EX_UNAVAILABLE
		EX_UNAVAILABLE  /* Give better error status */
#else
		1
#endif
		;	    
	}

	DPRINT(Debug,8,(&Debug,
			"mail: Exiting with exit status %d\n",
			Exstat));

	leave(0,Exstat);
    }

    if (0 == Exstat && !ret)
	Exstat = 1;

    if (exit_status) {
	*exit_status = Exstat;

	DPRINT(Debug,8,(&Debug,
			"mail: Setting exit status to %d\n",Exstat));
    }

    DPRINT(Debug,8,(&Debug,"mail=%d\n",ret));

    return ret;
}

/* Caller must call free_string() */
static struct string * exit_status_text P_((int exit_stat));
static struct string * exit_status_text(exit_stat)
     int exit_stat;
{
    struct string * ret = NULL;

#ifdef I_SYSEXITS

    /* Codes from sysexits.h */

    switch(exit_stat) {

#ifdef  EX_USAGE 
    case EX_USAGE    : 
	ret = format_string(CATGETS(elm_msg_cat, ElmSet,
				    ElmEX_USAGE,
				    "Usage error"));         break;
#endif
#ifdef  EX_DATAERR
    case EX_DATAERR  : 
	ret = format_string(CATGETS(elm_msg_cat, ElmSet,
				    ElmEX_DATAERR,
				    "Data format error"));   break;
#endif
#ifdef  EX_NOINPUT  
    case EX_NOINPUT  : 
	ret = format_string(CATGETS(elm_msg_cat, ElmSet,
				    ElmEX_NOINPUT,
				    "Cannot open input"));   break;
#endif
#ifdef  EX_NOUSER   
    case EX_NOUSER   : 
	ret = format_string(CATGETS(elm_msg_cat, ElmSet,
				    ElmEX_NOUSER,
				    "Address or user unknown")); break;
#endif
#ifdef  EX_NOHOST   
    case EX_NOHOST   : 
	ret = format_string(CATGETS(elm_msg_cat, ElmSet,
				    ElmEX_NOHOST,
				    "Host name unknown"));   break;
#endif
#ifdef  EX_UNAVAILABLE 
    case EX_UNAVAILABLE : 
	ret = format_string(CATGETS(elm_msg_cat, ElmSet,
				    ElmEX_UNAVAILABLE,
				    "Service unavailable")); break;
#endif
#ifdef  EX_SOFTWARE 
    case EX_SOFTWARE : 
	ret = format_string(CATGETS(elm_msg_cat, ElmSet,
				    ElmEX_SOFTWARE,
				    "Internal software error")); break;
#endif
#ifdef  EX_OSERR    
    case EX_OSERR    : 
	ret = format_string(CATGETS(elm_msg_cat, ElmSet,
				    ElmEX_OSERR,
				    "System error"));        break;
#endif
#ifdef  EX_OSFILE   
    case EX_OSFILE   : 
	ret = format_string(CATGETS(elm_msg_cat, ElmSet,
				    ElmEX_OSFILE   ,
				    "OS file missing"));     break;
#endif
#ifdef  EX_CANTCREAT
    case EX_CANTCREAT: 
	ret = format_string(CATGETS(elm_msg_cat, ElmSet,
				    ElmEX_CANTCREAT,
				    "Can't create output file")); break;
#endif
#ifdef  EX_IOERR    
    case EX_IOERR    : 
	ret = format_string(CATGETS(elm_msg_cat, ElmSet,
				    ElmEX_IOERR,
				    "Input/output error"));  break;
#endif
#ifdef  EX_TEMPFAIL 
    case EX_TEMPFAIL : 
	ret = format_string(CATGETS(elm_msg_cat, ElmSet,
				    ElmEX_TEMPFAIL,
				    "Temporary failure"));   break;
#endif
#ifdef  EX_PROTOCOL 
    case EX_PROTOCOL : 
	ret = format_string(CATGETS(elm_msg_cat, ElmSet,
				    ElmEX_PROTOCOL,
				    "Protocol error"));      break;
#endif
#ifdef  EX_NOPERM   
    case EX_NOPERM   : 
	ret = format_string(CATGETS(elm_msg_cat, ElmSet,
				    ElmEX_NOPERM,
				    "Permission denied"));   break;
#endif
#ifdef  EX_CONFIG   
    case EX_CONFIG   : 
	ret = format_string(CATGETS(elm_msg_cat, ElmSet,
				    ElmEX_CONFIG,
				    "Configuration error")); break;
#endif
    }    
#endif
    return ret;
}


static void mail_sent (fd,title,rs,ret,exit_code)
     union any_fd fd; 
     char * title;
     struct run_state *rs;
     int ret; 
     int exit_code;
{    
  lower_prompt(title);

  if (ret < 0) 
    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFailSignal,
		      "%.30s fail: Signal?"),
	      get_mailer_path(fd.mail_fd));
  else if (ret > 0) {
    if (rs->save_errno)
      lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFailErrno,
			"Failed: %.30s: %.40s"),
		get_mailer_path(fd.mail_fd),
		strerror(rs->save_errno));
    else if (exit_code) {
	struct string * sys_exit = exit_status_text(exit_code);

	if (sys_exit) {
	    lib_error(
		      CATGETS(elm_msg_cat, ElmSet, 
			      ElmMailerReturnedErrorT,
			      "mailer returned error status %d (%S)"), 
		      exit_code,sys_exit);
	    free_string(&sys_exit);
	} else
	    lib_error(
		      CATGETS(elm_msg_cat, ElmSet, 
			      ElmMailerReturnedError,
			      "mailer returned error status %d"), 
		exit_code);
    } else
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmMailSent, "Mail sent!"));
  } else {
      lib_error(CATGETS(elm_msg_cat, ElmSet, ElmLostErrno,
			"%.30s lost: %.40s"),
		get_mailer_path(fd.mail_fd),
		strerror(rs->save_errno));
  }
}

void dsn_menu P_((int *));
void dsn_menu(dsn)
     int * dsn;
{
  static int hdr_only = 0;
  static int failure  = 1;
  static int delay    = 1;
  static int success  = 0;
  static int DSN_off  = 0;

#define BOL 1
  static struct menu_item dsn_items[] = {
      { "Return H)eaders only on FAILURE:", 'h', 3,  BOL, 
	{ (char *) &hdr_only},
	sizeof hdr_only },
      { "Return DSN on F)AILURE         :", 'f', 5,  BOL, 
	{ (char *) &failure },
	sizeof failure },
      { "              D)ELAY           :", 'd', 6,  BOL, 
	{ (char *) &delay },
	sizeof delay },
      { "              S)UCCESS         :", 's', 7,  BOL, 
	{ (char *) &success },
	sizeof success },
      { "U)se defaults (DSN off)        :", 'u', 12, BOL, 
	{ (char *) &DSN_off },
	sizeof DSN_off }
  };

  if (0 == *dsn)
      DSN_off = 1;
  else {
    if (DSN_FULL    & *dsn) hdr_only = 0;
    if (DSN_HDRS    & *dsn) hdr_only = 1;
    if (DSN_SUCCESS & *dsn) success  = 1;
    if (DSN_FAILURE & *dsn) failure  = 1;
    if (DSN_DELAY   & *dsn) delay    = 1;
    if (DSN_NEVER   & *dsn) { success = 0; failure = 0; delay = 0; }
  }
  
  if (generic_menu(dsn_items,sizeof dsn_items / sizeof (struct menu_item),
	       "DSN (Delivery Status Notification) Configuration",
		   "DSN: ")) {

      if (DSN_off) *dsn = 0;
      else if (!success && !failure && !delay) *dsn = DSN_NEVER;
      else {
	  int val = 0;
	  if (hdr_only) val |= DSN_HDRS;
	  else          val |= DSN_FULL;
	  if (success)  val |= DSN_SUCCESS;
	  if (failure)  val |= DSN_FAILURE;
	  if (delay)    val |= DSN_DELAY;
	  
	  *dsn = val;
      }
  }
}

static int 	    edit_return_path(struct mailing_headers *headers,
				     struct menu_context *page);
static int 	    edit_return_path(headers, page)
     struct mailing_headers *headers;
     struct menu_context *page;
{

    if (headers->env_from) {
	int ret;
	int can_edit = 0;
	const char *t = mailer_env_from_value(headers->env_from,&can_edit);
	struct string *buffer = NULL;

	int LINES, COLUMNS;

	menu_get_sizes(page,&LINES, &COLUMNS);

	if (!can_edit) {
	    DPRINT(Debug,8, (&Debug,
			     "edit_return_path: Can not edit return path\n"));
	    return 0; /* bad command */
	}

	MoveCursor(LINES-4,0);
	CleartoEOS();
	       
	print_format_center(LINES-4, 
			    CATGETS(elm_msg_cat, ElmSet,
				    ElmEnterEnvFrom, 
				    "Enter value for envelope sender address."));

	if (t)
	    buffer = new_string2(display_charset,cs2cus(t));

	/* NOTE: Currently alias expansion is not possble */

	ret = optionally_enter2(page,
				&buffer,LINES-2, 0, 
				OE_REDRAW_MARK|
				OE_SIG_CHAR /* Ctrl-C */,
				CATGETS(elm_msg_cat, ElmSet, 
					ElmPreSEnvFromPrompt,
					"Return-path: "));
	
	if (ret == 0 && buffer) {
	    char * s = us2s(stream_from_string(buffer,0,NULL));

	    mailer_env_from_change(headers->env_from,s);

	    free(s);
	}
	
	if (buffer)
	    free_string(&buffer);
	
	if (ret == REDRAW_MARK)
	    return REDRAW_MARK;

	if (0 != ret)       /* Error or Ctrl-C */
	    return 1;

	ClearLine(5);

	t = mailer_env_from_value(headers->env_from,&can_edit);

	if (can_edit) {
	    if (t)
		PutLineX(12,0,CATGETS(elm_msg_cat, ElmSet, ElmPreSEnvFrom,
				     "R)eturn-path: %s"),
			 t);
	    else
		PutLineX(12,0,CATGETS(elm_msg_cat, ElmSet, ElmPreSEnvFromDef,
				     "R)eturn-path: (default)"));
	} else if (t)
	    PutLineX(12,0,CATGETS(elm_msg_cat, ElmSet, ElmPreSEnvFromNoEdit,
				 "Return-path: %s"),
		     t);

	return 1;
    }

    DPRINT(Debug,8, (&Debug,
		     "edit_return_path: env_from is not set\n"));
    return 0; /* bad command */
}

static enum pgp_confirm_loop {   /* jump targets */
    pcf_no   = 0,    /* No */
    pcf_out,         /* Yes */
    pcf_redraw,
    pcf_resize,
    pcf_EOF = -1
} pgp_send_confirm_loop P_((int replying, int was_pgp_encoded,
			    int *PGP_status, 
			    struct menu_context *page,
			    struct menu_context *prompt_area));
static enum pgp_confirm_loop pgp_send_confirm_loop(replying,was_pgp_encoded,
						   PGP_status,page,prompt_area)
     int replying; 
     int was_pgp_encoded;
     int *PGP_status;
     struct menu_context *page;
     struct menu_context *prompt_area;
{
    if (use_PGP &&
	PGP_status && replying && was_pgp_encoded &&
	! ((*PGP_status) & PGP_MESSAGE)) {

	menu_ClearScreen(prompt_area);

	menu_PutLineX(prompt_area,1, 0, 
		      CATGETS(elm_msg_cat, ElmSet, ElmPgpRcvSure, 
			      "The recv'd message was PGP encoded, are you sure? "));
	for (;;) {
	    int cmd = menu_ReadCh(prompt_area,REDRAW_MARK|READCH_resize);
	    
	    if (cmd == REDRAW_MARK) {
		DPRINT(Debug,4, (&Debug,"       .... redrawing\n"));
		menu_ClearScreen(page);   /* Clear possible redraw mark */
		
		/* Call refresh routines of children */
		menu_redraw_children(page);
		
		if (menu_need_redraw(prompt_area))		    
		    menu_ClearScreen(prompt_area);   /* Clear redraw mark from prompt_area*/
		
		return pcf_redraw;
	    }

	    if (cmd == RESIZE_MARK) {
		
		DPRINT(Debug,4, (&Debug,"       .... resizing\n"));
		return pcf_resize;
	    }

	    if (cmd == *def_ans_yes) {
		menu_Write_to_screen(prompt_area,
				     CATGETS(elm_msg_cat, ElmSet, ElmYesWord, 
					     "Yes."));
		return pcf_out;
	    }
	    if (cmd == *def_ans_no) {
		menu_Write_to_screen(prompt_area,
				     CATGETS(elm_msg_cat, ElmSet, ElmNoWord, 
					     "No."));
		return pcf_no;
	    }

	    if (cmd == EOF)
		return pcf_EOF;
	    
	}
    }
    
    return pcf_out;
}

static const char * have_ispell P_((void));
static const char * have_ispell() 
{
    const char * return_value   = ISPELL_path;

    if (! have_ISPELL || 
	! ISPELL_path[0]) {
	return_value = NULL;
    } else if (ISPELL_path[0] == '/') {
	if (-1 == access(ISPELL_path,EXECUTE_ACCESS)) {
            int err = errno;
            lib_error(CATGETS(elm_msg_cat, ElmSet,ElmCantIspell,
                              "Can't execute ispell: %s: %s"),
                      ISPELL_path,
                      strerror(err));
            DPRINT(Debug,6, (&Debug, 
                             "have_ispell: no access %s: %s\n",
                             ISPELL_path,
                             strerror(err)));
            sleep_message();
            return_value = NULL;
        }	
    }

    DPRINT(Debug,5,(&Debug,   
                    "have_ispell=%s\n",
                    return_value ? return_value : "<NULL>"));
    return return_value;

 }

/*
 * verify_transmission() - Ask the user to confirm transmission of the
 * message.  Returns 1 to send it,      -1 to forget it,
 *                   2 to preview,      -2 to forget empty file,
 *                   3 to preview menu, -3 to EOF on user input
 *                   4 to preview mime
 */
static enum verify_trans_result verify_transmission(filename, 
						    already_has_text, 
						    copy_file, force_cmd, 
						    options,
						    dsn, headers, attachments,
						    file_set,
						    mailer_info,
						    hdr_charset, mailbox,
						    encode_hdr,
						    aview,page,
						    prompt_area,
						    parent_message,
						    remote_server,
						    PGP_status,
						    snd_cmds)
     char *filename;	    /* pathname to mail mssg composition file        */
     int already_has_text;  /* indicates there is already text in the mssg   */
     struct copy_file *copy_file; /* pointer to buffer holding copy file name   */
     int force_cmd;	    /* command to do, '\0' to prompt user for cmd    */
     int options;
     int *dsn;
     struct mailing_headers *headers;   
     struct Attachments *attachments;
     charset_t file_set;      /* charset of file (filename) */
     struct mailer_info  *mailer_info;
     charset_t hdr_charset;
     struct MailboxView *mailbox;   /* needed for system_call() */
     int encode_hdr;
     struct AliasView *aview;
     struct menu_context *page;    /* new menu context assumed! */
     struct menu_context  *prompt_area;
     struct header_rec *parent_message;   /* Replying .... */
     const struct remote_server * remote_server;
     int * PGP_status;
     struct elm_commands *snd_cmds;
{
    struct string *prompt_mssg_s = NULL;	/* message to display prompting for cmd	*/
    struct string *prompt_menu1_s = NULL;       /* menu of available commands		*/
    struct string *prompt_menu2_s = NULL;
    int bad_cmd;		/* set TRUE to bitch about user's entry	*/
    int did_prompt;		/* TRUE if cmd prompted for and entered	*/
    int cmd;			/* command to perform			*/
    int x_coord, y_coord;
    int replying          = 0 != (options & MAIL_REPLYING);

    int was_pgp_encoded   = use_PGP &&
	0 != (options & MAIL_HAVE_PGP_ENCODED);

    int def_cmd = *def_ans_yes;
    int do_redraw = 1;
    int LINES, COLUMNS;
    enum user_level_v ul = 
	give_dt_enumerate_as_int(&user_level);
    
    enum verify_trans_result ret = ft_none;  
    int menu_build = 0;

    menu_get_sizes(page,&LINES, &COLUMNS);

    if (use_PGP &&
	PGP_status && replying && was_pgp_encoded &&
	! ((*PGP_status) & PGP_MESSAGE)) 
	def_cmd = 'P';

    bad_cmd = FALSE;		/* nothing to complain about yet	*/

    for (;;) {
	menu_set_default(page);

	if (0) {
	redraw:
	    do_redraw = 1;
	}

    resize_mark:
	if (menu_resized(page)) {
	    
	    menu_get_sizes(page,&LINES, &COLUMNS);

	    menu_subpage_relocate(prompt_area,page,LINES-4,4);

	    do_redraw = 1;

	} else if (menu_need_redraw(page))
	    do_redraw = 1;
	
	if (do_redraw) {

	    do_redraw = 0;
	    
	    menu_ClearScreen(page);
	    menu_StartXX(page,pg_BOLD);

	    menu_print_format_center(page,0, 
				     CATGETS(elm_msg_cat, ElmSet,
					     ElmMailScreenTitle, 
					     "Mail Pre-Send Screen"));

	    menu_EndXX(page,pg_BOLD);

	    show_last_error();

	    show_presend_headers(headers,hdr_charset, page);

	    if (headers->env_from && ul > 0) {
		int can_edit;
		const char *t = mailer_env_from_value(headers->env_from,&can_edit);

		if (can_edit) {
		    if (t)
			PutLineX(12,0,CATGETS(elm_msg_cat, ElmSet, ElmPreSEnvFrom,
					     "R)eturn-path: %s"),
				 t);
		    else
			PutLineX(12,0,CATGETS(elm_msg_cat, ElmSet, ElmPreSEnvFromDef,
					     "R)eturn-path: (default)"));
		} else if (t)
		    PutLineX(12,0,CATGETS(elm_msg_cat, ElmSet, ElmPreSEnvFromNoEdit,
					 "Return-path: %s"),
			     t);				     
	    }

	    menu_redraw_children(page);

	}

	/* build up prompt and menu strings */
	if (menu_build) {
	    ; /* not changed - no need to rebuild the strings */
	} else {

	    switch(ul) {
		
	    case user_level_beginner: {
		
		if (prompt_mssg_s)
		    free_string(&prompt_mssg_s);
		prompt_mssg_s = format_string(CATGETS(elm_msg_cat, ElmSet,
						      ElmVfyPromptSendTheMsg,
						      "Send the message now? y"));
		
		if (prompt_menu1_s)
		    free_string(&prompt_menu1_s);
		prompt_menu1_s = 
		    format_string(CATGETS(elm_msg_cat, ElmSet,
					  ElmVfyMenu1User0,
					  "Select letter of header to edit, 'e' to edit the message,"));

		if (prompt_menu2_s)
		    free_string(&prompt_menu2_s);
		prompt_menu2_s = 
		    format_string(CATGETS(elm_msg_cat, ElmSet,
					  ElmVfyMenu2User0V,
					  "'a' to make attachments, '%c' to send message, or '%c' to cancel"),
				  *def_ans_yes,*def_ans_no);
		
		/* In some conditions add these also to menu in user_level == 0 */
		
		if (query_mailer_info(mailer_info,MI_HAVE_DSN)) {
		    DPRINT(Debug,10,(&Debug,"mailer supports DSN\n"));
		    
		    if (dsn && *dsn != 0) {
			elm_append_message(&prompt_menu2_s,
					   CATGETS(elm_msg_cat, ElmSet,
						   ElmVfyMenuDSN0,
						   ", D)SN"));

		    }
		}

		if (use_PGP &&
		    replying && was_pgp_encoded) {

		    elm_append_message(&prompt_menu2_s,
				       CATGETS(elm_msg_cat, ElmSet,
					       ElmVfyMenuPGP0,
					       ", P)gp"));
		}

		elm_append_message(&prompt_menu2_s,FRM("."));
		
	    }
		break;
		
	    default: {
		if (prompt_mssg_s)
		    free_string(&prompt_mssg_s);
		prompt_mssg_s = format_string(CATGETS(elm_msg_cat, ElmSet,
						      ElmVfyPromptSendMessage,
						      "Send message now? y"));

		if (prompt_menu1_s)
		    free_string(&prompt_menu1_s);		
		prompt_menu1_s = 
		    format_string(CATGETS(elm_msg_cat, ElmSet,
					  ElmVfySelectLetter,
					  "Select letter of header to edit, "));

		elm_append_message(&prompt_menu1_s,
				   CATGETS(elm_msg_cat, ElmSet,
					   ElmVfyMenuEditMsg, 
					   "e)dit message, "));
		
		elm_append_message(&prompt_menu1_s,
				   CATGETS(elm_msg_cat, ElmSet,
					   ElmVfyMenuHeaders1, 
					   "all h)eaders, "));

		if (prompt_menu2_s)
		    free_string(&prompt_menu2_s);
		prompt_menu2_s = 
		    format_string(CATGETS(elm_msg_cat, ElmSet,
					  ElmVfyMenuVfyCpy,
					  "a)ttachments, co(p)y, "));

		if (have_ISPELL)
		    elm_append_message(&prompt_menu2_s,
				       CATGETS(elm_msg_cat, ElmSet, ElmVfyMenuIspell,
					       "i)spell, "));

		if (ALLOW_subshell)
		    elm_append_message(&prompt_menu2_s,
				       CATGETS(elm_msg_cat, ElmSet, ElmVfyMenuShell,
					       "!)shell, "));
		
		if (query_mailer_info(mailer_info,MI_HAVE_DSN)) {
		    DPRINT(Debug,10,(&Debug,"mailer supports DSN\n"));
		    
		    if (dsn)
			elm_append_message(&prompt_menu2_s,
					   CATGETS(elm_msg_cat, ElmSet,
						   ElmVfyMenuDSN,
						   "D)SN, "));
		}


		if (use_PGP) 
		    elm_append_message(&prompt_menu2_s,
				       CATGETS(elm_msg_cat, ElmSet,
					       ElmVfyMenuPGP,
					       "P)gp, "));

		if (mailbox) {
		    elm_append_message(&prompt_menu2_s,
				       CATGETS(elm_msg_cat, ElmSet,
					       ElmVfyMenuFolder,
					       "view) folder, "));
		}
		
		elm_append_message(&prompt_menu2_s,
				   CATGETS(elm_msg_cat, ElmSet,
					   ElmVfyMenuForget1,
					   "or f)orget"));		
	    }	   
		break;
	    }
	  
	    elm_append_message(&prompt_menu2_s,
			       CATGETS(elm_msg_cat, ElmSet,
				       ElmVfyMenuHelp,
				       " ? = help"));

    	    menu_build = 1;
	}

	if (bytes(filename) <= 0 &&
	    0 == attachments->attachment_count) {

	    if (prompt_mssg_s)
		free_string(&prompt_mssg_s);
	    prompt_mssg_s = format_string(CATGETS(elm_msg_cat, ElmSet,
						  ElmVfyPromptEmptySndMsg,
						  "Send empty message now? e"));

	    def_cmd = 'e';

	    menu_build = 0;  /* Rebuild menu */
	} else {
	    def_cmd = *def_ans_yes;

	    if (use_PGP &&
		PGP_status && replying && was_pgp_encoded &&
		! ((*PGP_status) & PGP_MESSAGE)) 
		def_cmd = 'P';
	}

	/* complain if last entry was bad */
	if (bad_cmd) {
	    menu_Write_to_screen(prompt_area,
				 FRM("%c??"), 07);
	    FlushBuffer();

	    if (sleepmsg > 0)
		error_sleep((sleepmsg + 1) / 2);
	    bad_cmd = FALSE;
	}

	if (menu_resized(prompt_area) ||
	    menu_need_redraw(prompt_area)) {
	
	    menu_ClearScreen(prompt_area);
	 
	    menu_PutLineX(prompt_area,
			  0, 0, FRM("%S"),prompt_mssg_s);

	    menu_print_center(prompt_area,1,prompt_menu1_s);
	    menu_print_center(prompt_area,2,prompt_menu2_s);

	    show_last_error();
	}

	/* if we don't have a cmd, display prompt and get response from user */
	if (force_cmd != '\0' &&
	    isascii(force_cmd)) {

	    cmd = force_cmd;     /* Removed tolower() from here */
		
	    force_cmd = '\0';
	    did_prompt = FALSE;

	    if (isascii(cmd) && 
		isprint(cmd)) {
		DPRINT(Debug,8,(&Debug,"... forced command: '%c'\n",
				cmd));
	    } else {
		DPRINT(Debug,8,(&Debug,"... forced command: %d\n",
				cmd));
	    }

	} else {
	    menu_PutLineX(prompt_area,
			  0, 0, FRM("%S"),prompt_mssg_s);

	    menu_GetXYLocation(prompt_area,
			       &x_coord, &y_coord);
	    y_coord--; /* backspace over default answer */

	    menu_CleartoEOLN(prompt_area);

	    menu_PutLineX(prompt_area,
			  0, y_coord, FRM("%c"),def_cmd);
	    FlushBuffer();

	    menu_MoveCursor(prompt_area,
			    x_coord, y_coord);
	    cmd = menu_ReadCh(prompt_area,REDRAW_MARK|READCH_resize);

	    if (isascii(cmd) && 
		isprint(cmd)) {
		DPRINT(Debug,8,(&Debug,"... command: '%c'\n",
				cmd));
	    } else {
		DPRINT(Debug,8,(&Debug,"... command: %d\n",
				cmd));
	    }
	   
	    if (cmd == REDRAW_MARK) {
		DPRINT(Debug,4, (&Debug,"       .... redrawing\n"));
		menu_ClearScreen(page);   /* Clear possible redraw mark */
		
		/* Call refresh routines of children */
		menu_redraw_children(page);

		if (menu_need_redraw(prompt_area))		    
		    menu_ClearScreen(prompt_area);   /* Clear redraw mark from prompt_area*/

		goto redraw;
	    }

	    if (cmd == RESIZE_MARK) {

		DPRINT(Debug,4, (&Debug,"       .... resizing\n"));
		goto resize_mark;
	    }

	    if (cmd == EOF)
		goto handle_EOF;

	    did_prompt = TRUE;
	    clear_error();

	    /* Put cursor pack to correct place */
	    menu_MoveCursor(prompt_area,
			    x_coord, y_coord);
	}

	switch (cmd) {
	case '\n':
	case '\r':
	    cmd = def_cmd;

	    if (isascii(cmd) && 
		isprint(cmd)) {
		DPRINT(Debug,8,(&Debug,"... default command: '%c'\n",
				cmd));
	    } else {
		DPRINT(Debug,8,(&Debug,"... default command: %d\n",
				cmd));
	    }
	}

	if (cmd == *def_ans_yes) {
	    if (did_prompt) {
	        menu_Write_to_screen(prompt_area,
				     CATGETS(elm_msg_cat, ElmSet, ElmVfyMenuSend,
					     "Send"));
		/* Call FlushBuffer() after writing of "Send" so that
		 * user get immediately feedback (it may take time
		 * to encode big attachments)
		 */
		FlushBuffer();
	    }

	    switch(pgp_send_confirm_loop(replying,was_pgp_encoded,
					 PGP_status,page,prompt_area)) {
	    case pcf_no: break;
	    case pcf_out: 
		ret = ft_send;
		goto out;
	    case pcf_redraw: goto redraw;
	    case pcf_resize: goto resize_mark;
	    case pcf_EOF:   goto handle_EOF;
	    }


	} else if (cmd == *def_ans_no )
	    goto do_no;

	/* handle command */
	switch (cmd) {
	
	case ' ':
	    if (did_prompt) {
	        menu_Write_to_screen(prompt_area,
				     CATGETS(elm_msg_cat, ElmSet, ElmVfyMenuPreview,
					     "Preview message"));
		FlushBuffer();
	    }

	    switch(pgp_send_confirm_loop(replying,was_pgp_encoded,
					 PGP_status,page,prompt_area)) {
	    case pcf_no: break;
	    case pcf_out: 
		ret = ft_preview;
		goto out;
	    case pcf_redraw: goto redraw;
	    case pcf_resize: goto resize_mark;
	    case pcf_EOF:   goto handle_EOF;
	    }
	    break;

	case '!':
	    menu_ClearLine(prompt_area,0);
	    if (ALLOW_subshell)
		subshell(NULL, page, prompt_area);
	    else
		lib_error(CATGETS(elm_msg_cat, ElmSet, 
				  ElmNoSubshellCommand,
				  "Subshell command isn't configured in this version of ELM."));
	      
	    if (menu_need_redraw(page)) {
		menu_ClearScreen(page);

		do_redraw = 1;
	    }
	    break;

	 case '?':
	     if (did_prompt) {
		 menu_Write_to_screen(prompt_area,
				      FRM("?"));
		 FlushBuffer();
	     }
	    
	    {		
		int ch = help_generic(snd_cmds,TRUE,page,prompt_area);
				
		if (menu_need_redraw(page)) {
		    menu_ClearScreen(page);
		    
		    do_redraw = 1;
		}
		
		if (EOF == ch) 
		    goto handle_EOF;
	    }
	    break;

	case 'a':
	    if (did_prompt) {
	        menu_Write_to_screen(prompt_area,
				     CATGETS(elm_msg_cat, ElmSet, ElmVfyMenuAtttach,
					     "Attachments"));
		FlushBuffer();
	    }

	    attach_menu(NULL,NULL,attachments,display_charset,mailer_info,
			NULL, aview,mailbox,NULL,remote_server);

	    do_redraw = 1;
	    break;

        case 'D':
	    if (dsn) {
		
		if (did_prompt) {
		    menu_Write_to_screen(prompt_area,
					 CATGETS(elm_msg_cat, ElmSet, ElmVfyMenuDsn,
						 "Dsn"));
		    FlushBuffer();
		}
		
		if (query_mailer_info(mailer_info,MI_HAVE_DSN)) {
		    DPRINT(Debug,8,(&Debug,"mailer supports DSN\n"));
		    
		    dsn_menu(dsn);
		    do_redraw = 1;
		} else {
		    lib_error(CATGETS(elm_msg_cat, ElmSet, 
				      ElmVfyMenuDsnNotSupported,
				      "Mailer does not support delivery notification settings."));
		    
		}
	    } 
	    break;

	case ctrl('D') :
	    if (dsn) {

		if (did_prompt) {
		    menu_Write_to_screen(prompt_area,
					 CATGETS(elm_msg_cat, ElmSet, ElmVfyMenuDsnToggle,
						 "Dsn toggle"));
		    FlushBuffer();
		}
		
		if (query_mailer_info(mailer_info,MI_HAVE_DSN)) {

		    if (0 == *dsn || 
			0 != (*dsn & DSN_NEVER) ||
			0 == (*dsn & DSN_SUCCESS)) {

			*dsn &=  ~DSN_NEVER;
			*dsn |=   DSN_SUCCESS|DSN_FAILURE|DSN_DELAY;
		
			lib_transient(CATGETS(elm_msg_cat, ElmSet, 
					      ElmVfyMenuDsnEnabled,
					      "Delivery notifications enabled.")); 

		    } else if (0 != (*dsn & DSN_SUCCESS)) {
			*dsn &=   ~DSN_SUCCESS;

			if (0 == *dsn) {
			    *dsn = DSN_NEVER;
			    lib_transient(CATGETS(elm_msg_cat, ElmSet, 
						  ElmVfyMenuDsnDisabled,
						  "Delivery notifications disabled."));
			
			} else
			    lib_transient(CATGETS(elm_msg_cat, ElmSet, 
						  ElmVfyMenuDsnSucceesDisabled,
						  "Delivery notification for SUCCESS disabled."));
			
		    } 
		    
		} else {
		    lib_error(CATGETS(elm_msg_cat, ElmSet, 
				      ElmVfyMenuDsnNotSupported,
				      "Mailer does not support Dsn notification settings."));
		    
		}		    
	    }
	    break;

	case 'e': 
	    if (did_prompt) {
		menu_Write_to_screen(prompt_area,
				     CATGETS(elm_msg_cat, ElmSet, 
					     ElmVfyMenuEdit, "Edit"));
		FlushBuffer();
	    }

	    /* can trigger redraw 
	       return 0 on OK */
	    if (0 != edit_the_message(filename, already_has_text,
				 headers,
				 NULL /* use "editor" or "alteditor" */,
				 file_set,
				 mailer_info, mailbox,
				 aview, page)) {
		ret = ft_none;
		goto out;
	    }
	    break;

	case ctrl('F'):
	    if (did_prompt) {
		menu_Write_to_screen(prompt_area,
				     CATGETS(elm_msg_cat, ElmSet,
					     ElmPassphraseForget,
					     "Forget passphrase"));
		FlushBuffer();
	    }

	    forget_passphrase();	      
	    break;

	  case 'f':
	    do_no:
	      if (did_prompt) {
		  menu_Write_to_screen(prompt_area,
				       CATGETS(elm_msg_cat, ElmSet, ElmVfyMenuForget, 
					       "Forget"));
		  FlushBuffer();
	      }

	      if (PGP_status)
		  *PGP_status= 0; /* make sure to reset! */
	      
	      if (bytes(filename) <= 0 &&
		  0 == attachments->attachment_count) {
		  ; /* forget about empty files */
		  ret = ft_forget_empty;
		  goto out;
	      } 
	      ret = ft_forget;
	      goto out;
	      /*NOTREACHED*/

	case 'H':
	    if (did_prompt) {
	        menu_Write_to_screen(prompt_area,
				     CATGETS(elm_msg_cat, ElmSet, ElmVfyMenuPreviewHdrs,
					     "Preview message with headers"));
		FlushBuffer();
	    }

	    switch(pgp_send_confirm_loop(replying,was_pgp_encoded,
					 PGP_status,page,prompt_area)) {
	    case pcf_no: break;
	    case pcf_out: 
		ret = ft_preview_hdrs;
		goto out;
	    case pcf_redraw: goto redraw;
	    case pcf_resize: goto resize_mark;
	    case pcf_EOF:   goto handle_EOF;
	    }
	    break;

	case 'h':
	    if (did_prompt) {
		menu_Write_to_screen(prompt_area,
				     CATGETS(elm_msg_cat, ElmSet, 
					     ElmVfyMenuHeaders,"Headers"));
		FlushBuffer();
	    }

	    edit_headers(headers,mailer_info,hdr_charset,encode_hdr,
			 aview,parent_message);
	    
	    do_redraw = 1;
	    break;

	case 'i': 
	    if (did_prompt) {
		menu_Write_to_screen(prompt_area,
				     CATGETS(elm_msg_cat, ElmSet, 
					     ElmVfyMenuIspell2,"Ispell"));
		FlushBuffer();
	    }


	    if (have_ISPELL) {

		const char * have_it = have_ispell();

		if (have_it) {
		    char * lbuf = elm_message(FRM("%s %s %s"),
					      have_it, ISPELL_options, filename);
		    Raw(OFF);
		    system_call(lbuf, SY_ENAB_SIGHUP, NULL);
		    
		    do_redraw = 1;
		    Raw(ON);

		    free(lbuf);
		}
	    } else
		lib_error(CATGETS(elm_msg_cat, ElmSet, 
				  ElmNoIspellCommand,
				  "Ispell command isn't configured in this version of ELM."));
	
	    break;

	case ctrl('L'):
	    goto redraw;

	case 'p':
	    if (did_prompt) {
		menu_Write_to_screen(prompt_area,
				     CATGETS(elm_msg_cat, ElmSet, 
					     ElmVfyMenuCopyFile, 
					     "Copy file"));
		FlushBuffer();
	    }
	    name_copy_file(copy_file, aview, mailbox,page);
	    menu_trigger_redraw(prompt_area);   
	    break;

        case 'P':
	    if (did_prompt) {
	        menu_Write_to_screen(prompt_area,
				     CATGETS(elm_msg_cat, ElmSet, 
					     ElmVfyMenuPgp,"Pgp"));
		FlushBuffer();
	    }
	    
	    if (use_PGP) {
		if (PGP_status && !*PGP_status) {
		    *PGP_status = pgp_menu (filename,headers, page);
		    if (*PGP_status)
			def_cmd = *def_ans_yes;
		    else
			def_cmd = 'P';
		    do_redraw = 1;

		} else
		    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmPgpAlreadyEncSig,
				      "This message is already encrypted and/or signed!"));
	    } else
		lib_error(CATGETS(elm_msg_cat, ElmSet, 
				  ElmNoPgpConfigured,
				  "PGP isn't configured in this version of ELM."));
            break;

	case 'R':
	    {
		int c = edit_return_path(headers, page);

		DPRINT(Debug,8,(&Debug,"edit_return_path returned %d\n",c));

		if (0 == c)
		    bad_cmd = TRUE;
		else if (REDRAW_MARK == c)
		    goto redraw;		    

		menu_trigger_redraw(prompt_area);
	    }
	    break;

	case 'v': 
	    if (did_prompt) {
	        menu_Write_to_screen(prompt_area,
				     CATGETS(elm_msg_cat, ElmSet, ElmVfyMenuPrevMime,
					     "Preview parts of message"));
		FlushBuffer();
	    }

	    switch(pgp_send_confirm_loop(replying,was_pgp_encoded,
					 PGP_status,page,prompt_area)) {
	    case pcf_no: break;
	    case pcf_out: 
		ret = ft_preview_mime;
		goto out;
	    case pcf_redraw: goto redraw;
	    case pcf_resize: goto resize_mark;
	    case pcf_EOF:   goto handle_EOF;
	    }

	    break;
		

	case 'V':
	    if (did_prompt) {
	        menu_Write_to_screen(prompt_area,
				     CATGETS(elm_msg_cat, ElmSet, ElmVfyMenuPrevMenu,
					     "Preview menu"));
		FlushBuffer();
	    }
	    switch(pgp_send_confirm_loop(replying,was_pgp_encoded,
					 PGP_status,page,prompt_area)) {
	    case pcf_no: break;
	    case pcf_out: 
		ret = ft_preview_menu;
		goto out;
	    case pcf_redraw: goto redraw;
	    case pcf_resize: goto resize_mark;
	    case pcf_EOF:   goto handle_EOF;
	    }

	    break;

	case 'w':
	    if (mailbox) {
		int ch;

		if (did_prompt) {
		    menu_Write_to_screen(prompt_area,
					 CATGETS(elm_msg_cat, ElmSet, 
						 ElmVfyMenuViewFolder,
						 "View folder"));
		    FlushBuffer();
		}

		ch = send_view_folder(mailbox, attachments, page, aview,
				 mailer_info);
		
		if (menu_need_redraw(page)) {
		    menu_ClearScreen(page);

		    do_redraw = 1;
		}

		if (EOF == ch) 
		    goto handle_EOF;

	    } else
		bad_cmd = TRUE;
	    break;
    
	default:

	    {
		int c = presend_action(headers,mailer_info,cmd,hdr_charset,
				       aview, page, prompt_area,
				       parent_message);

		if (0 == c) {   /* Accept also uppercase letters */
		    int old = cmd;
#ifdef ASCII_CTYPE
		    if (isascii(cmd))
#endif
			cmd = tolower((unsigned char)cmd);
		    
		    if (old != cmd)
			c = presend_action(headers,mailer_info,cmd,hdr_charset,
					   aview, page, prompt_area, parent_message);

		    if (0 == c)
			bad_cmd = TRUE;
		    else if (REDRAW_MARK == c)
			goto redraw;		    
		    
		} else if (REDRAW_MARK == c)
		    goto redraw;		 

		menu_trigger_redraw(prompt_area);   
	    }
	    
	    break;
	}
    }

    if (0) {	
    handle_EOF:
	DPRINT(Debug,2,(&Debug,"EOF on verify_transmission\n"));
	
	ret = ft_EOF;
    }

 out:

    if (prompt_menu2_s)
	free_string(&prompt_menu2_s);
    if (prompt_menu1_s)
	free_string(&prompt_menu1_s);
    if (prompt_mssg_s)
	free_string(&prompt_mssg_s);

    DPRINT(Debug,10,(&Debug,"verify_transmission="));
    switch (ret) {
    case ft_none: DPRINT(Debug,10,(&Debug,"ft_none")); break;
    case ft_send: DPRINT(Debug,10,(&Debug,"ft_send")); break;
    case ft_preview: DPRINT(Debug,10,(&Debug,"ft_preview")); break;
    case ft_preview_menu: DPRINT(Debug,10,(&Debug,"ft_preview_menu")); break;
    case ft_preview_mime: DPRINT(Debug,10,(&Debug,"ft_preview_mime")); break;
    case ft_preview_hdrs: DPRINT(Debug,10,(&Debug,"ft_preview_hdrs")); break;
    case ft_forget: DPRINT(Debug,10,(&Debug,"ft_forget")); break;
    case ft_forget_empty: DPRINT(Debug,10,(&Debug,"ft_forget_empty")); break;
    case ft_EOF: DPRINT(Debug,10,(&Debug,"ft_EOF")); break;
    }
    DPRINT(Debug,10,(&Debug," (%d)\n",ret));

    return ret;
}

void write_header_info(mailer, headers,
		       copy, mime_info)
     struct out_state *mailer;
     struct mailing_headers * headers;  
     enum cm_across_mode   copy;
     mime_send_t *mime_info;
{
    /**
       
        Added the ability
	to have backquoted stuff in the users .elmheaders file!
	If copy is TRUE, then treat this as the saved copy of outbound
	mail.
    **/
    

    mime_info -> encode_hdr = !allow_no_hdrencoding;
	
    /** Subject moved to top of headers for mail because the
	pure System V.3 mailer, in its infinite wisdom, now
	assumes that anything the user sends is part of the 
	message body unless either:
	1. the "-s" flag is used (although it doesn't seem
	   to be supported on all implementations? )
	2. the first line is "Subject:".  If so, then it'll
	   read until a blank line and assume all are meant
	   to be headers.
	So the gory solution here is to move the Subject: line
	up to the top.  I assume it won't break anyone elses program
	or anything anyway (besides, RFC-822 specifies that the *order*
	of headers is irrelevant).  Gahhhhh....
    **/

    if (headers->subject)
	write_string_header(mailer,"Subject",
			    headers->subject,
			    mime_info->encoding_top,
			    mime_info->encode_hdr,
			    mime_info->hdr_charset);

    {
	const char *MIME_name_h = 
	    get_charset_MIME_name(mime_info->hdr_charset);


	/* Our standard violations */
	char OSV[100];
	OSV[0] = '\0';
	  
	if (!mime_info->encode_hdr && MIME_name_h)
	    add_parameter(OSV,"hdr-charset",
			  MIME_name_h,sizeof(OSV),FALSE);
	
	/* If allow_no_hdrencoding is set, there is no mime-encoding on
	   user headers either
	*/
	if (allow_no_hdrencoding) {
	    add_parameter(OSV,"no-hdr-encoding","1",sizeof(OSV),FALSE);	    
	}

	if (OSV[0]) {
	    state_printf(mailer, 
			 FRM("X-ELM-OSV: (Our standard violations) %s"),
			 OSV);
	    print_EOLN(mailer,mime_info->encoding_top);
	}
    } 
    
    if (headers->in_reply_to)
	write_references_header(mailer,"In-Reply-To",
				headers->in_reply_to,
				mime_info->encoding_top,
				mime_info->encode_hdr,
				mime_info->hdr_charset);
    if (headers->references)
	write_references_header(mailer,"References",
				headers->references,
				mime_info->encoding_top,
				mime_info->encode_hdr,
				mime_info->hdr_charset);


    if (headers->to.addrs)
	write_addr_header(mailer,"To",headers->to.addrs,
			  mime_info->encoding_top,
			  mime_info->encode_hdr,
			  mime_info->hdr_charset,1);
    write_text_header(mailer,"Date",get_arpa_date(),
		      mime_info->encoding_top);

    if (headers->sender)
	write_text_header(mailer,"Sender",headers->sender,
			  mime_info->encoding_top);
	
    if (headers->from.addrs) 
	write_addr_header(mailer,"From",headers->from.addrs,
			  mime_info->encoding_top,
			  mime_info->encode_hdr,
			  mime_info->hdr_charset,0);
    
    if (headers->cc.addrs)
	write_addr_header(mailer,"CC",headers->cc.addrs,
			  mime_info->encoding_top,
			  mime_info->encode_hdr,
			  mime_info->hdr_charset,1);

    if ((copy != cm_across_send && headers->bcc.addrs) ||
	/* If there is no To or CC recipients,
	 * show BCC recipients to make mail legal 
	 */
	(!headers->to.addrs && !headers->cc.addrs && headers->bcc.addrs))
	write_addr_header(mailer,"Bcc",headers->bcc.addrs,
			  mime_info->encoding_top,
			  mime_info->encode_hdr,
			  mime_info->hdr_charset,1);
	
    if (headers->action)
	write_text_header(mailer,"Action",headers->action,
			  mime_info->encoding_top);

    if (headers->priority)
	write_text_header(mailer,"Priority",headers->priority,
			  mime_info->encoding_top);

    if (headers->precedence)
	write_text_header(mailer,"Precedence",headers->precedence,
			  mime_info->encoding_top);

    if (headers->expires)
	write_text_header(mailer,"Expires",headers->expires,
			  mime_info->encoding_top);

    if (headers->reply_to.addrs)
	write_addr_header(mailer,"Reply-To",headers->reply_to.addrs,
			  mime_info->encoding_top,
			  mime_info->encode_hdr,
			  mime_info->hdr_charset,1);
	
    if (headers->user_header_count > 0) {
	int i;
	for (i = 0; i <  headers->user_header_count; i++) {
	    write_string_header(mailer,
				give_header_name(headers->user_header[i].
						 name),
				headers->user_header[i].value,
				mime_info->encoding_top,
				mime_info->encode_hdr,
				mime_info->hdr_charset);
	}
    }

#ifndef NO_XHEADER
    state_printf(mailer, FRM("X-Mailer: ELM [version %s]"), version_buff);
    print_EOLN(mailer,mime_info->encoding_top);
#endif /* !NO_XHEADER */

    mime_write_top_headers (mailer, mime_info);

    mime_info->cl_offset = 0;
    if (copy == cm_across_copy) {
	const char * t = NULL;
	int x;

	if (headers->env_from)
	    t = mailer_env_from_value(headers->env_from,&x);

	if (t)
	    write_text_header(mailer,"Return-path",t,
			      mime_info->encoding_top);
	
	/* Only write content-length to copy */
	state_puts("Content-Length: ",mailer);
	mime_info->cl_offset = out_state_ftell(mailer);
	state_puts("          ",mailer); /* Print Blanks as Placeholders */
	print_EOLN(mailer,mime_info->encoding_top);
	state_puts("Status: RO",mailer);
	print_EOLN(mailer,mime_info->encoding_top);
    }
    print_EOLN(mailer,mime_info->encoding_top);
}

static int copy_part P_((mime_send_t *mime_info,struct out_state *dest,
			 FILE *conv_file,struct mime_send_part * X,
			 int copy));
static int copy_part(mime_info,dest,conv_file,X,copy)
     mime_send_t *mime_info;
     struct out_state *dest;
     FILE *conv_file;
     struct mime_send_part * X;
     int copy;
{
    int line_len;
    char buffer[SLEN];			/* file reading buffer */
    long a UNUSED_VAROK = out_state_ftell(dest);
    long pos1;

    if (0 != fseek(conv_file,X->start_loc,SEEK_SET)) {
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmSeekFailed,
			  "ELM [seek] failed trying to read %d bytes into file."),
		  (int) X->start_loc);
	return 0;
    }

       
    while ((pos1 = ftell(conv_file)) < X->end_loc &&
	   (line_len = mail_gets(buffer, sizeof buffer -1, conv_file))) {

	/* Text part necessary do not end to \n so there is possibility to
	   get part of next part 
	*/

	if (pos1 + line_len > X->end_loc) {
	    int old = line_len;

	    line_len = X->end_loc - pos1;

	    DPRINT(Debug,4,(&Debug,"copy_part: Got past of end of attachment -- changing line_len %d -> %d\n",
			    old,line_len));
	    if (line_len < 0 || line_len > old)
		mime_panic(__FILE__,__LINE__,
			   "copy_part",
			   "line_len calculation error");

	    if (0 == line_len)
		break;

	}


	if (X->encoding_part == ENCODING_BINARY) {
	    /* \n -> \r\n conversion is done for that part already */
	    state_put(buffer, line_len, dest);
	} else if (mime_info->encoding_top == ENCODING_BINARY) {
	    /* It is better perhaps use canonical eol (CRLF) when mail have
	     * content transfer encoding BINARY somewhere (see notes about 
	     * BINARYMIME)
	     */
	    
	    /* THIS assumes that 
	       dest->EOLN_is_CRLF is set
	    */
	    state_convert_EOLN(buffer,&line_len,sizeof buffer,dest);

	    /* We don't do escaping because line does not end either 
	     * to plain \n now...
	     */
	    state_put(buffer, line_len, dest);
#ifndef DONT_ESCAPE_MESSAGES
	} else if (copy && (strncmp(buffer, "From ", 5) == 0)) {
	    /* Add in the > to a From on our copy */
	    state_puts(">",dest);
	    state_put(buffer, line_len, dest);
#endif
#ifdef NEED_LONE_PERIOD_ESCAPE
	} else if (!copy && strcmp(buffer, ".\n") == 0) {
	    /* Because some mail transport agents take a lone period to
	     * mean EOF, we add a blank space on outbound message.
	     */
	    state_puts(". \n", dest);
#endif /* NEED_LONE_PERIOD_ESCAPE */
	} else {
	    /* SMTP uses \r\n */
	    if (out_state_EOLN_is_CRLF(dest))
		state_convert_EOLN(buffer,&line_len,sizeof buffer,dest);

	    state_put(buffer, line_len, dest);
	}
    }

    if (out_state_ferror(dest) || ferror(conv_file))
	return 0;

    DPRINT(Debug,4, (&Debug,
		     "Preparing mail for %s: part %d -- %d bytes written to temp%s\n",
		     copy ? "copy" : "sending",
		     X-mime_info->top_parts,
		     out_state_ftell(dest)-a,
		     feof(conv_file) ? ", got EOF" : ""));

    return 1;
}

int copy_message_across(mime_info, dest, copy, conv_file)
     mime_send_t *mime_info;
     struct out_state *dest;
     enum cm_across_mode copy;
     FILE *conv_file;
{
    char * mode UNUSED_VAROK = "????";

    /** If copy is TRUE, treat as a saved copy of outbound mail. **/

    mime_info -> cl_start = out_state_ftell(dest);

    switch(copy) {
    case cm_across_send:    mode = "sending";   break;
    case cm_across_copy:    mode = "copy";      break;
    case cm_across_preview: mode = "preview";   break;
	
    }

    if (mime_info->msg_is_multipart) {
	int i;

	DPRINT(Debug,4, (&Debug,
			 "Preparing mail for %s: %d parts (multipart)\n",
			 mode,
			 mime_info->top_parts_count));

	for (i = 0; i < mime_info->top_parts_count; i++) {
	    if (cm_across_copy != copy || 
		mime_info->top_parts[i].save_it_on_copy) {

		print_EOLN(dest,mime_info->encoding_top);
		state_printf(dest,FRM("--%s"),mime_info->mime_boundary);
		print_EOLN(dest,mime_info->encoding_top);

		mime_write_part_headers(dest,mime_info,
					&(mime_info->top_parts[i]));
		print_EOLN(dest,mime_info->encoding_top);

		if (!copy_part(mime_info,dest,conv_file,
			       &(mime_info->top_parts[i]),copy)) {
		    return 0;
		}
	    }
	}
	print_EOLN(dest,mime_info->encoding_top);
	state_printf(dest,FRM("--%s--"), mime_info->mime_boundary);
	print_EOLN(dest,mime_info->encoding_top);

    } else if (1 == mime_info->top_parts_count) {
	DPRINT(Debug,4,(&Debug,  
		  "Preparing mail for %s: one part (not multipart)\n",
		  mode));

	if (cm_across_copy != copy || 
	    mime_info->top_parts[0].save_it_on_copy) {
	    if (!copy_part(mime_info,dest,conv_file,
			   &(mime_info->top_parts[0]),copy)) {
		return 0;
	    }
	}
    } else {
	DPRINT(Debug,4,(&Debug,  
			"Preparing mail for %s: ODD: %d parts (NOT multipart)\n",
			mode,mime_info->top_parts_count));
    }

    mime_info->cl_end = out_state_ftell(dest) ;
    if (mime_info->cl_offset > 0) {
	/* go fixup the content length header */
	out_state_fseek(dest, mime_info->cl_offset);
	state_printf(dest, FRM("%ld"), 
		     (long)(mime_info->cl_end - mime_info->cl_start));
	  
	/* Return to the end of the file! */
	out_state_fseek (dest, mime_info->cl_end);
    }

    DPRINT(Debug,4,(&Debug,  
		    "Preparing mail for %s: total %d bytes body, ~%d bytes headers\n",
		    mode,
		    mime_info->cl_end - mime_info -> cl_start,
		    mime_info -> cl_start));
    
    /* !! */
    if (EOF == fflush(out_state_FILE(dest)) || ferror(out_state_FILE(dest))) {
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmWriteFailedCopy,
			  "Write failed to temp file in copy"));
	return 0;
    }

    return 1;
}

static int have_remote_address P_((struct addr_list * addr_list));
static int have_remote_address(addr_list)
     struct addr_list * addr_list;
{
    int idx;
    int len = addr_list_item_count(addr_list);

    const char **mailname_list = give_dt_path_as_elems(&mailname,"mailname");
    
    /* check To: list */
    
    for (idx = 0; idx < len; idx++) {
	
	int group = -1;
	const struct address * address = 
	    addr_list_get_item(addr_list,idx,&group);
	int is_local = 0;
	
	const char  * ad  = address_get_ascii_addr(address);
	
	const char *ptr;

	if (!ad)
	    continue;
	
	ptr = qstrpbrk_c(ad,":@");
	
	/* Have route, so address is remote */
	if (ptr && ':' == *ptr)
	    return 1;

	/* Address do not have '@' character, 
	   so this address is local         */
	if (!ptr)
	    continue;

	ptr++;
	
	if (mailname_list) {
	    
	    int i;
	    
	    for (i = 0; mailname_list[i]; i++) {
		
		if (0 == istrcmp(mailname_list[i],ptr)) {
		    DPRINT(Debug,12,(&Debug,
				     "have_remote_address: %s is local [mailname[%d]=%s]\n",
				     ptr,i,mailname_list[i]));
		    
		    is_local = 1;
		    break;
		}
	    }

	} else {   /* mailname setting overrides hostname and hostfullname */

	    is_local = istrcmp(ptr, hostfullname) == 0;

	    if (is_local) {
		DPRINT(Debug,12,(&Debug,
				 "have_remote_address: %s is local\n",
				 ptr));
	    }
	}

	if (! is_local)
	    return 1;
	
    }

    return 0;
}

static int append_sig(file, headers, page)
     FILE *file;
     struct mailing_headers *headers;
     struct menu_context *page;
{
    /* Append the correct signature file to file.  Return TRUE if
       we append anything.  */
    
    /* Look at the to and cc list to determine which one to use */
    
    /* We could check the bcc list too, but we don't want people to
       know about bcc, even indirectly */
    
    /* Some people claim that  user@anything.same_domain should be 
       considered local.  Since it's not the same machine, better be 
       safe and use the remote sig (presumably it has more complete
       information).  You can't necessarily finger someone in the
       same domain. */

    const char * lsig = give_dt_estr_as_str(&local_signature_e, 
					    "localsignature",NULL,NULL);
    const char * rsig = give_dt_estr_as_str(&remote_signature_e,
					    "remotesignature",NULL,NULL);


    if (!lsig || !rsig)
	error_wait();

    /* FIXME -- is this correct? */
    if (! OPMODE_IS_BATCHMODE(opmode) && (lsig || rsig)) {

	char filename2[SLEN];
	const char *sig = "";

	/* check each @ for @thissite.domain */
	/* if any one is different than this, then use remote sig */

	sig = lsig;

	if (headers->to.addrs &&
	    have_remote_address(headers->to.addrs))
	    sig = rsig;	   	    
	

	/* still local? */         
	if (sig == lsig &&
	    headers->cc.addrs &&
	    have_remote_address(headers->cc.addrs))
	    sig = rsig;
    
	
	if (sig && sig[0]) {  /* if there is a signature file */
	    if (sig[0] != '/')
		elm_sfprintf(filename2, sizeof filename2,
			     FRM("%s/%s"), 
			     home, sig);
	    else
		strfcpy(filename2, sig, sizeof filename2);

	    /* append the file - with a news 2.11 compatible */
	    /* seperator if "sig_dashes" is enabled */
	    (void) append(file, filename2, (sig_dashes ? "\n-- \n" : NULL),
			  page);
	    
	    return TRUE;
	}    
    }
    return FALSE;
}

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
