static char rcsid[] = "@(#)$Id: span_range.c,v 1.9 2016/09/28 18:29:37 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.9 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@siilo.FMI.FI>
 *      or  Kari Hurtta <elm@elmme-mailer.org>
 *****************************************************************************/

#include "def_pager.h"

DEBUG_VAR(Debug,__FILE__,"pager");

static enum span_result span_range_prefix P_((struct menu_context *ctx,
					      int cur_line, int *cur_col,
					      uint16 first_char,
					      struct pager_range  * pager_range,
					      int *this_flags,
					      int *mayclear,
					      int wrap_indicator));

static enum span_result span_range_prefix(ctx,cur_line,cur_col,
					  first_char, pager_range, this_flags,
					  mayclear,wrap_indicator)
     struct menu_context *ctx;
     int cur_line; 
     int *cur_col;
     uint16 first_char;
     struct pager_range  * pager_range;
     int *this_flags;
     int *mayclear;     
     int wrap_indicator;
{
    enum span_result res = span_ok;

    int CUR_col    = *cur_col;
    int THIS_flags = *this_flags;
    int MAY_clear = *mayclear;

    int START_col UNUSED_VAROK = CUR_col;

    int range_flags = get_pager_range_flags(pager_range);
    int quote_level = get_pager_range_quote_level(pager_range);  /* Original flowed quote */
    int indent      = get_pager_range_indent(pager_range);
    
    int WP =  /* pager_indicate_wrapping */ wrap_indicator;

    int LINES, COLUMNS;

    menu_get_sizes(ctx, &LINES, &COLUMNS);

    DPRINT(Debug,12,(&Debug,
		     "span_range_prefix: line = %d, col = %d, first_char=%04x, LINES=%d, COLUMNS=%d\n",
		     cur_line,CUR_col,first_char, LINES, COLUMNS));

    DPRINT(Debug,12,(&Debug,
		     "span_range_prefix: range_flags = %d, quote_level = %d, indent = %d\n",
		     range_flags,quote_level,indent));

    if (0 != (range_flags & PR_REPLY_QUOTE))                     /* Should not happen */
	quote_level++;

    if (quote_level && quote_prefix) {
	int lastX = -1, i;
	uint16 repchar = 0x0;
	int POS = 0;

	int len = string_len(quote_prefix);

	DPRINT(Debug,15,(&Debug,
			 "span_range_prefix: (len=%d) prefix=%S\n",
			 len,quote_prefix));

	for (i = 0; i < len; i++) {
	    uint16 ch  = give_unicode_from_string(quote_prefix,i);

	    if (0 != unicode_ch(ch,UOP_noctrl) &&
		0 == unicode_ch(ch,UOP_space)) {
		repchar = ch;
		lastX = i;
	    }
	}
	
	if (len > 0)
	    first_char = give_unicode_from_string(quote_prefix,0);

	if (0 != (range_flags & PR_QUOTE_L)) {                  /* Should not happen */
	    int need_quote = 0x005B /* [ */ == first_char;
	    
	    DPRINT(Debug,12,(&Debug,
			     "span_range_prefix: need_quote=%d first_char=%04x\n",
			     need_quote,first_char));

	    menu_StartXX(ctx,pg_BOLD);
	    if (need_quote)
		menu_Writechar(ctx,'[');
	    menu_EndXX(ctx,pg_BOLD);

	}

	if (lastX >= 0) {
	    DPRINT(Debug,12,(&Debug,
			     "span_range_prefix: lastX=%d, repchar=%04x\n",
			     lastX,repchar));

	    res = span_line(ctx,cur_line,&CUR_col,quote_prefix,&POS,&MAY_clear,
			    THIS_flags,COLUMNS/3,lastX+1,WP);

	    if (res != span_ok || POS <= lastX) {
		DPRINT(Debug,12,(&Debug,
				 "span_range_prefix: failure, pos=%d, col=%d, res=%d\n",
				 POS,CUR_col,res));
		goto failure;
	    }
	} else {
	    DPRINT(Debug,12,(&Debug,
			     "span_range_prefix: No printable characters on prefix\n"));
	}

	if (quote_level > 1 && repchar) {
	    int x;

	    if (THIS_flags)
		menu_StartXX(ctx,THIS_flags);


	    for (x = 0; x < quote_level - 1 && CUR_col < COLUMNS/2; x++) {
	
		menu_WriteUnicode(ctx,repchar);
		CUR_col++;       
	    }
	
	    if (THIS_flags)
		menu_EndXX(ctx,THIS_flags);

	}


	if (POS < len) {
	    res = span_line(ctx,cur_line,&CUR_col,quote_prefix,&POS,&MAY_clear,
			    THIS_flags,COLUMNS/3,len-lastX,WP);
	    
	    if (res != span_ok || POS < len) {
		DPRINT(Debug,12,(&Debug,
				 "span_range_prefix: failure, pos=%d, col=%d, res=%d\n",
				 POS,CUR_col,res));
		goto failure;
	    }
	    
	}

    failure:
	;

    } else if (0 != (range_flags & PR_QUOTE_L)) {            /* Should not happen */
	int need_quote = 0x005B /* [ */ == first_char;
	
	DPRINT(Debug,12,(&Debug,
			 "span_range_prefix: need_quote=%d first_char=%04x\n",
			 need_quote,first_char));
	
	menu_StartXX(ctx,pg_BOLD);
	if (need_quote)
	    menu_Writechar(ctx,'[');
	menu_EndXX(ctx,pg_BOLD);
	
    }

    while (CUR_col < COLUMNS - COLUMNS/3 && 
	   CUR_col < indent) {	
	menu_WriteUnicode(ctx,0x0020 /* SPACE */ );
	CUR_col++;       
    }

    *cur_col    = CUR_col;
    *this_flags = THIS_flags;
    *mayclear   = MAY_clear;

    DPRINT(Debug,12,(&Debug,
		     "span_range_prefix=%d (col = %d -> %d, line = %d)\n",
		     res,START_col, CUR_col, cur_line));

    return res;
}

static enum span_result span_range_middle P_((struct menu_context *ctx,
					      int cur_line, int *cur_col,
					      uint16 first_char,
					      struct pager_range  * pager_range,
					      int *this_flags,
					      int *mayclear));

static enum span_result span_range_middle(ctx,cur_line,cur_col,
				   first_char, pager_range, this_flags,
				   mayclear)
     struct menu_context *ctx;
     int cur_line; 
     int *cur_col;
     uint16 first_char UNUSED_VAROK;
     struct pager_range  * pager_range;
     int *this_flags;
     int *mayclear;     
{
    enum span_result res = span_ok;

    int CUR_col    = *cur_col;
    int THIS_flags = *this_flags;
    int MAY_clear = *mayclear;

    int START_col UNUSED_VAROK = CUR_col;

    int range_flags UNUSED_VAROK = get_pager_range_flags(pager_range);
    int quote_level UNUSED_VAROK = get_pager_range_quote_level(pager_range);  /* Original flowed quote */
    int indent      = get_pager_range_indent(pager_range);

    int LINES, COLUMNS;

    menu_get_sizes(ctx, &LINES, &COLUMNS);

    DPRINT(Debug,12,(&Debug,
		     "span_range_middle: line = %d, col = %d, first_char=%04x, LINES=%d, COLUMNS=%d\n",
		     cur_line,CUR_col,first_char, LINES, COLUMNS));

    DPRINT(Debug,12,(&Debug,
		     "span_range_middle: range_flags = %d, quote_level = %d, indent = %d\n",
		     range_flags,quote_level,indent));

        while (CUR_col < COLUMNS - COLUMNS/3 && 
	   CUR_col < indent) {	
	menu_WriteUnicode(ctx,0x0020 /* SPACE */ );
	CUR_col++;       
    }

    *cur_col    = CUR_col;
    *this_flags = THIS_flags;
    *mayclear   = MAY_clear;

    DPRINT(Debug,12,(&Debug,
		     "span_range_middle=%d (col = %d -> %d, line = %d)\n",
		     res,START_col, CUR_col, cur_line));

    return res;
}

const struct span_helper_data NULL_span_helper_data = {
    SPAN_HELPER_DATA_magic,

    0,0,0,NULL,0,0,0,0,0,0,0,0,0,0,0
};

enum span_result span_helper (ctx,pager_range,info,buffer,data)
     struct menu_context * ctx;
     struct pager_range  * pager_range;
     struct pg_wrap_info * info;
     struct string       * buffer;
     struct span_helper_data * data;
{
    enum span_result span_line_result;

    int L_mayclear      = 0;

    int buffer_len      = string_len(buffer);
    int LINES, COLUMNS;
    int WP;

    if (SPAN_HELPER_DATA_magic != data->magic)
	panic("PAGER PANIC",__FILE__,__LINE__,
	      "span_helper",
	      "Bad magic number",0);

    L_mayclear          = data->mayclear;
    WP  =  /* pager_indicate_wrapping */ data->wrap_indicator;

    menu_get_sizes(ctx, &LINES, &COLUMNS);

    if (pager_range) {
	data->range_flags = get_pager_range_flags(pager_range);
	
	DPRINT(Debug,9,(&Debug, 
			"span_helper: pager range: quote_level=%d range_flags=%d\n",
			get_pager_range_quote_level(pager_range),
			data->range_flags));
    }

    if (data->start_of_line) {
	DPRINT(Debug,10,(&Debug, 
			 "span_helper: virtual_line=%d, start of line\n",
			 data->virtual_line));
	
	record_pg_wrap_info(info,&(data->virtual_line),data->idx,
			    data->buffer_x,data->lineext);
    }

    if (data->start_of_line && pager_range) {
	uint16 ch = 0;   /* First character may be needed */
	
	if (buffer_len > 0)
	    ch = give_unicode_from_string(buffer,0);
	
	span_line_result = span_range_prefix(ctx,
					     data->cur_line,&(data->cur_col),
					     ch,pager_range,
					     &(data->this_flags),
					     &L_mayclear,
					     WP
					     );
	
	DPRINT(Debug,9,(&Debug, 
			"span_helper: span_range_prefix result = %d, ",
			span_line_result));
	
	data->start_of_line = 0;
	data->p_start_COL   = data->cur_col;
	
	if (data->start_of_range) {
	    DPRINT(Debug,9,(&Debug,"start of range (range changed) ignored, "));
	    data->start_of_range = 0;
	}

	if (data->buffer_x >= buffer_len) {
	    DPRINT(Debug,9,(&Debug, 
			    "no buffer left (buffer_x=%d buffer_len=%d), ",
			    data->buffer_x,buffer_len));
	    goto buffer_handled;
	}
	
    } else if (data->start_of_range && pager_range) {

	uint16 ch = 0;   /* First character may be needed */
	
	if (buffer_len > 0)
	    ch = give_unicode_from_string(buffer,0);

	span_line_result = span_range_middle(ctx,
					     data->cur_line,&(data->cur_col),
					     ch,pager_range,
					     &(data->this_flags),
					     &L_mayclear);
	
	DPRINT(Debug,9,(&Debug, 
			"span_helper: span_range_middle result = %d, ",
			span_line_result));

	data->start_of_range = 0;
	
	if (data->buffer_x >= buffer_len) {
	    DPRINT(Debug,9,(&Debug, 
			    "no buffer left (buffer_x=%d buffer_len=%d), ",
			    data->buffer_x,buffer_len));
	    goto buffer_handled;
	}

    } else {
	if (data->start_of_line) {
	    DPRINT(Debug,10,(&Debug, 
			     "span_helper: No pager range on beginning of line\n"));
	    
	    data->start_of_line = 0;
	    data->p_start_COL   = 0;
	}
	
	if (data->start_of_range) {
	    DPRINT(Debug,10,(&Debug, 
			     "span_helper: start of range (range changed) without range\n"));
	    
	    data->start_of_range = 0;
	}


	if (data->joined_line) {
	    DPRINT(Debug,10,(&Debug, 
			     "span_helper: this is joined line\n"));
	}

	if (0 != (data->range_flags & PR_CENTER_THIS) &&
	    COLUMNS > buffer_len &&
	    data->p_start_COL  == data->cur_col) {

	    /* Does not work correctly with double with characters 
	       also does not work with control characters
	     */

	    int pos = (COLUMNS - buffer_len) / 2;
	  
	    if (pos > data->cur_col) {
		DPRINT(Debug,10,(&Debug, 
				 "span_helper: centering this line, col=%d => %d\n",
				 data->cur_col,pos));

		menu_MoveCursor(ctx,data->cur_col,pos);

		/* Make sure that garbage from previous
		   output is not left */
		if (L_mayclear) {
		    menu_CleartoEOLN(ctx);
		    L_mayclear = 0;
		}
		
		menu_MoveCursor(ctx,data->cur_line,pos);
		data->cur_col     = pos;
		data->p_start_COL = pos;
	    }	    	   
	}

	
	if (0 != (data->range_flags & PR_WORD_WRAP) 
	    || 
	    (data->force_word_wrap && 0 == (data->range_flags & PR_PREFORMAT))
	    ) {
	    
	    span_line_result = span_words(ctx,
					  data->cur_line,&(data->cur_col),
					  buffer,&(data->buffer_x),
					  &L_mayclear,
					  data->this_flags,
					  COLUMNS    /* max_width*/,
					  buffer_len /* max_len */,
					  data->joined_line,
					  data->p_start_COL,
					  0 == (data->range_flags & 
						PR_MAX_WIDTH) ? 
					  data->p_width : 0,
					  WP);
	    
	    DPRINT(Debug,9,(&Debug, 
			    "span_helper: span_words result = %d, ",
			    span_line_result));
	    
	    
	} else {
	    
	    /* This is the part of the code that actually displays on 
	     * the screen 
	     */
	    
	    span_line_result = span_line(ctx,
					 data->cur_line,&(data->cur_col),
					 buffer,&(data->buffer_x),
					 &L_mayclear,
					 data->this_flags,
					 COLUMNS    /* max_width*/,
					 buffer_len /* max_len */,
					 WP);
	    DPRINT(Debug,9,(&Debug, 
			    "span_helper: span_line result = %d, ",
			    span_line_result));
	    
	}
	
    buffer_handled:
	if (data->lineext) {
	    DPRINT(Debug,9,(&Debug," line extension (joining), "));
	    
	    data->joined_line = 1;
	} else if (0 == (data->range_flags & PR_JOIN_LINES)) {
	    DPRINT(Debug,9,(&Debug," line wrapping, "));
	    data->start_of_line = 1;  /* Need new line */
	} else {
	    DPRINT(Debug,9,(&Debug," joining lines, "));
	    data->joined_line = 1;
	}
    }
    
    data->mayclear      = L_mayclear ? 1 : 0;

    DPRINT(Debug,9,(&Debug," span_helper=%d | ",span_line_result));

    /* pager_helper_process_line() on src/pager/helper.c prints rest of debug line */

    return span_line_result;
}

void clear_span_helper_data(data)
     struct span_helper_data *data;
{
    if (SPAN_HELPER_DATA_magic != data->magic)
	panic("PAGER PANIC",__FILE__,__LINE__,
	      "clear_span_helper_data",
	      "Bad magic number",0);

    if (data->lineext)
	free_pager_lineext(&(data->lineext));
}


/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
